/*********************************************************************
 *   _      _        _     _                             _           *
 *  | |__  (_) _ __ | | __| |  ___  _ __    _ __    ___ | |_         *
 *  | '_ \ | || '__|| |/ /| | / _ \| '__|  | '_ \  / _ \| __|        *
 *  | |_) || || |   |   < | ||  __/| |   _ | | | ||  __/| |_         *
 *  |_.__/ |_||_|   |_|\_\|_| \___||_|  (_)|_| |_| \___| \__|        *
 *                                                                   *
 *********************************************************************
 * Copyright 2003                                                    *
 * Jrgen Birkler (birkler@yahoo.com)                                *
 *********************************************************************/

/**
 * @file 
 *
 * Main accelR8 code
 *
 * @author Jrgen Birkler (birkler@yahoo.com)
 * 
 */


#include "hardware.h"
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include "printf_P.h"
#include "debug.h"
#include "timers.h"
//#include "monitor.h"
#include "display.h"
#include "i2c.h"
#include "adxl.h"
#include "rtc.h"
#include "buzzer.h"
#include "isqrt.h"
#include "monitor.h"
 

const ADXL_CalibrationInfo_t EEPROM_SavedCalibrationData __attribute__((section(".eeprom"))) = ADXL_CALIBRATION_INITIAL_VALUE;


///Current LCD bias voltage in mV
static volatile Voltage_t LCD_Voltage_Current = 0;

///Desired LCD bias voltage in mV
static volatile Voltage_t LCD_Voltage_Set = -1900;

///Current charge pump timer
static Milliseconds_t LCD_ChargePumpTimer = 100;

static void UpdateLCDChargePumpTimer(void);


///Period between LCD bias voltage measurements
#define TIMER_READ_ADC_TIMEOUT_SHORT 200
#define TIMER_READ_ADC_TIMEOUT_LONG 1000


///Timeout if on before switch off
#define TIMER_SWITCH_OFF_ON_TIMEOUT 60000

//Timeout between display updates
#define TIMER_UPDATE_DISPLAY_TIMEOUT 250 //ms


/**
 * Allocate CPU clock
 * Do NOT power save, go to idle, keep clock running
 */
static void CPU_Clock_Allocate(void);

///Free the CPU clock so that powersave is possible
static void CPU_Clock_Free(void);

///Status of CPU clock allocations. @return !0 if CPU clock is allocated
static unsigned char CPU_Clock_IsAllocated(void);

///Get the current MUX setting. @returns 0..7 if ADC coversion is complete. -1 if no conversion is complete
static int8_t ADC_GetMUX(void);

///Latest voltage read from ADC
static Voltage_t voltage;

///Latest value read from ADC
static volatile unsigned int ADC_ReadValue;


///Time it takes for the OSC2 TIMER2 oscillator to stabilize. 1s according to spec.
#define TIMER_TIMER2_STABILIZE_TIME 1000


///Da uart send char used by #uart_printf
void uart_sendchar(char data, void* not_used);

///Wrapper for uart printfs
#define uart_printf(format, args...)   _formatted_write(PSTR(format),uart_sendchar,0, ## args)


#if _SFR_ASM_COMPAT==1
#error "Skit"
#endif //_SFR_ASM_COMPAT

typedef enum {
  AccelR8_Mode_Off,
  AccelR8_Mode_FIRST,
  AccelR8_Mode_Measure = AccelR8_Mode_FIRST,
  AccelR8_Mode_LateralAcceleration,
  AccelR8_Mode_Calibrate,
  AccelR8_Mode_SwitchOff,
  AccelR8_Mode_LAST
} AccelR8_Mode_t;

/*
void eeprom_write_block(const void *buf, unsigned int addr, size_t n)
{
  const uint8_t* p = (const uint8_t*)buf;
  while (n-- > 0)
  {
    while (!eeprom_is_ready())
      ;
    eeprom_wb(addr++,*(p++));
  }
}
*/

static void Mode_Off_Enter(void)
{
  ADXL_PowerOff();
  CPU_Clock_Free();
  DEBUG_PRINT_INFO_("\r\noff.");
  Display_Off();
  LCD_CHARGE_PUMP_OFF();
  Timer_Reset(TIMER_READ_ADC);
  Timer_Reset(TIMER_SWITCH_OFF);
  Timer_Reset(TIMER_UPDATE_DISPLAY);
  Timer_Reset(TIMER_LCD_CHARGE_PUMP);
  Timer_Reset(TIMER_LOG);

}

static void Mode_Off_Leave(void)
{

  ADXL_PowerOn();
  CPU_Clock_Allocate();

  Display_On();

  //Let charge pump timer and ADC measurement expired immediatly
  Timer_SetExpired(TIMER_LCD_CHARGE_PUMP,1);
  Timer_SetExpired(TIMER_READ_ADC,TIMER_READ_ADC_TIMEOUT_SHORT);     

  Timer_SetExpired(TIMER_UPDATE_DISPLAY,TIMER_UPDATE_DISPLAY_TIMEOUT);

  Timer_SetExpired(TIMER_LOG,100);

}

static AccelR8_Mode_t currentMode = AccelR8_Mode_Off;

static void Mode_NewMode(AccelR8_Mode_t newMode)
{
  //Leave current mode
  switch (currentMode)
  {
    case AccelR8_Mode_Off:
      Mode_Off_Leave();
      break;

    case AccelR8_Mode_LAST:
      break;

    default:
      break;

  }

  //Set new mode
  currentMode = newMode;

  //Enter new mode
  switch (currentMode)
  {
    case AccelR8_Mode_Off:
      Mode_Off_Enter();
      break;

    case AccelR8_Mode_Measure:
    case AccelR8_Mode_LateralAcceleration:
      {
        ADXL_CalibrationInfo_t info;
        eeprom_read_block(&info, (void*)&EEPROM_SavedCalibrationData, sizeof(info));
        ADXL_SetCalibrationInfo(&info);
      }
      printf("\f");
      Timer_SetExpired(TIMER_UPDATE_DISPLAY,TIMER_UPDATE_DISPLAY_TIMEOUT);
      break;

    case AccelR8_Mode_SwitchOff:
      Timer_Set(TIMER_SWITCH_OFF,3000);
      printf("\f");
      Timer_SetExpired(TIMER_UPDATE_DISPLAY,TIMER_UPDATE_DISPLAY_TIMEOUT);
      break;

    case AccelR8_Mode_Calibrate:
      ADXL_ResetMeasuredCalibrationInfo();
      Timer_SetExpired(TIMER_UPDATE_DISPLAY,TIMER_UPDATE_DISPLAY_TIMEOUT);
      break;

    default:
    case AccelR8_Mode_LAST:
      printf("\f");
      Timer_SetExpired(TIMER_UPDATE_DISPLAY,TIMER_UPDATE_DISPLAY_TIMEOUT);
      break;
  }
}


static unsigned long int z_offset = 0;

static unsigned long int current_acceleration = 0;

static signed long int horiz_speed = 0;



static void UpdateDisplay(void)
{
  switch (currentMode)
  {
    case AccelR8_Mode_Measure:
    case AccelR8_Mode_LateralAcceleration:
      {
        ADXL_g_t axisX = ADXL_GetX();
        ADXL_g_t axisY = ADXL_GetY();
        //const ADXL_CalibrationInfo_t* info_p =
        //  ADXL_GetMeasuredCalibrationInfo();

        //ADXL_TimerTicks_t scale = info_p->Tp;
        axisX/=10;
        axisY/=10;

        current_acceleration = sqr_li(axisX);
        current_acceleration += sqr_li(axisY);

        if (currentMode == AccelR8_Mode_Measure)
        {
          printf("\vX-Y\r\nx:%+6dmg\r\ny:%+6dmg",axisX,axisY);
        }
        else {
          ADXL_g_t a_abs;// = lisqrt(current_acceleration);
          signed long int temp = current_acceleration - z_offset;
          ADXL_g_t a_rel;
          if (temp <= 0)
            a_rel = 0;
          else
            a_rel = lisqrt(temp);

          temp = a_rel * TIMER_UPDATE_DISPLAY_TIMEOUT;

          if (axisX < 0)
          {
            temp = -temp;
          }

          if (axisY < 0)
          {
            temp -= temp;
          }

          horiz_speed += temp;

          a_abs = horiz_speed * 98L / 10000L;

          printf("\vSPD/ACC\r\nv:%+6dmg\r\na:%+6dmg",a_abs,a_rel);
        }
      }
      break;


    case AccelR8_Mode_Calibrate:
      {
        //Print value from ADXL
        const ADXL_CalibrationInfo_t* info_p =
          ADXL_GetMeasuredCalibrationInfo();

        ADXL_CalibrationInfo_t temp = *info_p;

        //ADXL_g_t axisX = ADXL_GetX();
        //ADXL_g_t axisY = ADXL_GetY();
        //ADXL_TimerTicks_t scale = ADXL_GetAverage(); //ADXL_GetScale();
        //ADXL_TimerTicks_t scale = ADXL_GetScale();

        printf("\vCalibrate\r\n%5d %5d\r\n%5d %5d",temp.TxMax,temp.TxMin,temp.TyMax,temp.TyMin);
      }
      break;

    case AccelR8_Mode_SwitchOff:
      {
        //Update RTC
        static uint8_t prevSec = 0xff;
        const Time_t* time_p = RTC_Get();
        
        if (prevSec != time_p->seconds)
        {
          printf("\vOff...\r\n%02d:%02d:%02d",(int)time_p->hour,(int)time_p->minute,(int)time_p->seconds);
        }
      }
      break;

    default:
      break;
  }
}










/**
 * Da main program
 *
 * @todo Keep track of on/off state
 * @todo User interface
 * @todo Handle RESET and MODE key presses
 * @todo __naked__ attribute
 *
 */
int main(void)
{
  //Initialize
  //////////////////////
  //CPU_INIT Must be first since it enabled all input port's pull resistors to avoid
  //power consumption from undefined voltages to schmitt triggers
  CPU_INIT();
  I2C_Init();
  ADC_INIT();
  Display_Init();
  Display_On();
  LEDS_INIT();
  BLUE_LED_ON();
  RED_LED_ON();
  UART_INIT(9600);
  ADXL_INIT();
  RTC_Init();
  Timers_Init();
  WAKEUP_INIT();
  BUTTONS_INIT();

  //Startup
  ///////////////////////////////////////////

  //TRAP();

  //Reserve clock until async timer clock has stabilized
  CPU_Clock_Allocate();
  Timer_Set(TIMER_TIMER2_STABILIZE,TIMER_TIMER2_STABILIZE_TIME);

  //Start to blink the alive LED
  Timer_Set(TIMER_BLUE_LED,500);


  //Enable interrupts
  ///////////////////////////  
  sei();

  //We're alive!
  DEBUG_PRINT_INFO_("\r\nAVR up and running...");
  DEBUG_PRINT_INFO_("\r\naccelR8 (C) 2002 Jrgen Birkler birkler@yahoo.com\r\n");
  Buzzer_Buzz();

  //Leave off state
  Mode_NewMode(AccelR8_Mode_FIRST);

  //Main program loop begin
  /////////////////////////////
  FOREVER 
  {
    static bool ButtonPressed = false;

    //Update the LCD charge pump timer
    UpdateLCDChargePumpTimer();

    if (ButtonPressed)
    {
      if (!BUTTON_RESET_DOWN() && !BUTTON_MODE_DOWN())
      {
        ButtonPressed = false;
      }
    }
    else {
      if (BUTTON_MODE_DOWN())
      {
        AccelR8_Mode_t mode = currentMode;
        
        ButtonPressed = true;
        
        Buzzer_Buzz();

        mode++;
        if (mode >= AccelR8_Mode_LAST) mode = AccelR8_Mode_FIRST;

        Timer_Set(TIMER_SWITCH_OFF,TIMER_SWITCH_OFF_ON_TIMEOUT);

        Mode_NewMode(mode);
      }

      if (BUTTON_RESET_DOWN())
      {
        ButtonPressed = true;

        Buzzer_Buzz();

        switch (currentMode)
        {
          case AccelR8_Mode_Calibrate:
            {
              const ADXL_CalibrationInfo_t* info_p =
                ADXL_GetMeasuredCalibrationInfo();

              eeprom_write_block(info_p,(void*)&EEPROM_SavedCalibrationData,sizeof(ADXL_CalibrationInfo_t));
            }
            break;
          case AccelR8_Mode_LateralAcceleration:
            z_offset = current_acceleration;
            horiz_speed = 0;
            break;

          default:
            break;

        }
      }
    }

    //Update display
    if (Timer_HasExpired(TIMER_UPDATE_DISPLAY)) 
    {
      UpdateDisplay();
    }

    if (Timer_HasExpired(TIMER_SWITCH_OFF)) 
    {
      Mode_NewMode(AccelR8_Mode_Off);
    }

    if (Timer_HasExpired(TIMER_LOG))
    {
      ADXL_g_t axisX = ADXL_GetX();
      ADXL_g_t axisY = ADXL_GetY();

      axisX/=10;
      axisY/=10;

      uart_printf("%+05d %+05d\r", axisX, axisY);
    }

    //Read ADC
    if (Timer_HasExpired(TIMER_READ_ADC)) 
    {
      dbg_printf("\r\nVlcd:%6dmV tmr:%6dms", LCD_Voltage_Current, LCD_ChargePumpTimer);

      ADC_ReadValue = 0xFFFF;
      CPU_Clock_Allocate();
      ADC_ENABLE();
      ADC_SETUP_MUX(ADC_MUX_VLCD);
      ADC_ENABLE_INTERRUPT();
      ADC_START();
    } 

    //LCD Charge Pump
    if (Timer_HasExpired(TIMER_LCD_CHARGE_PUMP)) 
    {
      LCD_CHARGE_PUMP_TOGGLE();
      Timer_Set(TIMER_LCD_CHARGE_PUMP,LCD_ChargePumpTimer);     
    }

    //Wait 1s for 32KHz crystal to stabilize before going to powersave mode
    if (Timer_HasExpired(TIMER_TIMER2_STABILIZE)) 
    {
      Timer_Reset(TIMER_TIMER2_STABILIZE);
      DEBUG_PRINT_INFO_("\r\nTIMER2 stable...\r\n");
      CPU_Clock_Free();
    }

    //LED blink alive indication
    if (Timer_HasExpired(TIMER_BLUE_LED)) 
    {
      BLUE_LED_CHANGE();
    }

    //Check if any timer already expired
    if (Timers_GotoToIdle())
    {
      //Is clock needed by any resource?
      if (CPU_Clock_IsAllocated() || !Buzzer_PowerSaveOk())
      {
        //Yes. Only goto idle
        //CPU_IDLE();

                                              
        MCUCR |= _BV(SE);                          
        __asm__ __volatile__ ("sleep" "\n\t" :: ); 
        MCUCR &= ~_BV(SE);                         
     }
      else {
        //Nope, no clock needed. Powersave.
        RED_LED_OFF();
        UART_POWER_SAVE_SLEEP();
        ADC_DISABLE();
        CPU_POWER_SAVE();
        UART_POWER_SAVE_WAKEUP();
        RED_LED_ON();
      }
    }
  }
  //Main program loop end
}


static void UpdateLCDChargePumpTimer(void)
{
  //If mux != -1 then ADC value is available in 'voltage'
  int8_t mux = ADC_GetMUX();

  if (mux == ADC_MUX_VLCD)
  {
    Voltage_t prev = LCD_Voltage_Current;
    LCD_Voltage_Current = ADC_LCD_V_LCD(voltage);
    static Voltage_t delta = 0;

    Voltage_t diff = LCD_Voltage_Set - LCD_Voltage_Current;

    if (diff < 500)
      Timer_Set(TIMER_READ_ADC, TIMER_READ_ADC_TIMEOUT_LONG);

    delta = LCD_Voltage_Current - prev;

    diff /= 20;
    delta /= 10;

    //New charge pump timer based on current LCD voltage
    LCD_ChargePumpTimer+=diff - delta;


    if (LCD_ChargePumpTimer > 5000) LCD_ChargePumpTimer = 20;
  }
}


void sendchar(char c)
{
  Display_OutChar(c);
  //UART_WRITE_BYTE(c);
}


void dbg_sendchar(char data, void* not_used)
{
  not_used = not_used;
  Monitor_SendChar(data);
  //UART_WRITE_BYTE(data);
}


void uart_sendchar(char data, void* not_used)
{
  not_used = not_used;
  //Monitor_SendByte(data);
  UART_WRITE_BYTE(data);
}


//ADC
//////////////////////////


static volatile int8_t adc_mux = -1;

INTERRUPT(SIG_ADC)
{
  //Read mux
  unsigned int value = ADC_GET_VALUE(); 
  Voltage_t v = ADC_MEASUREMENT_TO_VOLTAGE(value);
  ADC_ReadValue = value;
  voltage = v;

  adc_mux = ADC_GET_MUX();

  CPU_Clock_Free();
  ADC_DISABLE_INTERRUPT();
  ADC_DISABLE();
}


static int8_t ADC_GetMUX(void)
{
  int8_t temp = adc_mux;
  adc_mux = -1;
  return temp;
}


INTERRUPT(SIG_INTERRUPT1)
{
}




///CPU Clock
/////////////////////

///Number of CPU clock allocations
static volatile unsigned char cpu_clock_allocs = 0; 


static void CPU_Clock_Allocate(void)
{
  cli();
  cpu_clock_allocs++;
  sei();
}

static void CPU_Clock_Free(void)
{
  cli();
  cpu_clock_allocs--;
  sei();
}

static unsigned char CPU_Clock_IsAllocated(void)
{
  return cpu_clock_allocs;
}



/**
 * @mainpage Bullit Accelerometer
 *
 * The accelerometer can measure acceleration in two axis
 * using Analog Devices ADXL202. +-2g can be measured in each axis.
 * Look \ref ADXL_References_AnalogMems "here" for datasheets and application notes.
 *
 * \image HTML pcb.jpg "Accelerometer with microcontroller and display"
 *
 * \section ADXL_MainPage_Intro Introduction
 * - \ref ADXL_MainPage_Features - Features and functions planned
 * - \ref ADXL_MainPage_TheoryOfOperation - Explains the software and calculations from the sensor
 * - \ref ADXL_MainPage_Hardware - Hardware overview and design description.
 *   Download \ref ADXL_MainPage_Download "Schematic and PCB design" can be downloaded. Shematic
 *   and PCB is drawn with <a href="http://www.cadsoft.de/">Eagle Layout Program</a>.
 * - \ref ADXL_MainPage_CodePieces - Describes some of the code pieces that are part of the software.
 *   All \ref ADXL_MainPage_Download "source code" can be downloaded.
 * - \ref ADXL_MainPage_Download - Download source code and hardware schematics
 *
 * \section ADXL_MainPage_Features Features
 * @copydoc ADXL_UserManual_Features
 * See the \ref ADXL_UserManual_Page.
 *
 * \section ADXL_MainPage_Hardware Hardware
 * @copydoc ADXL_Hardware_Page
 * Each hardware is wrapped by a macro in the
 * hardware.h file. No port is accessed directly without using a macro.
 * @copydoc ADXL_HardwareAssignments
 *
 * \section ADXL_MainPage_TheoryOfOperation Theory of Operation
 * @ref ADXL_TheoryOfOperation
 *
 * <A HREF="http://www.futurama.com">...magic. Got it.</A>
 *
 * \section ADXL_MainPage_Building Building the code
 * Install the \ref ADXL_References_Cygwin "CYGWIN" AVR-GCC \ref ADXL_References_Tools "development environment"
*  for Windows.
 * This is my \ref ADXL_Makefile_Page with the auto configuration of the RTC clock
 * to start counting the time from when the code was built. The excellent 
 * \ref ADXL_References_Doxygen is used to generate the documentation.
 *
 * \section ADXL_MainPage_Libraries Libraries and Utilities
 * The accelerometer consists of the following software components:
 * - \ref ATMEL_AVR_Embedded_TimerUtilities
 *   Timers are used as unexact timing of display updates, LCD charge pumps and debug blinks
 * - \ref ATMEL_AVR_Embedded_RTC  creates the timebase of the accelerometer.
 *   It also keeps and internal clock complete with date and time.
 * - \ref ATMEL_AVR_Embedded_I2C used by the \ref ATMEL_AVR_Embedded_Display
 * - \ref ATMEL_AVR_Embedded_FIRFilter used by measurement utilities to
 *   filter period time and measurements.
 * - \ref ADXL_. Uses input capture of TIMER1 to measure
 *   PWM output from ADXL202.
 *
 * \section ADXL_MainPage_CodePieces Software Components
 * The code consist of a number of parts that run concurrently. A number of
 * interrupts and the main loop.
 *
 * \subsection ADXL_MainPage_Timer1_Interrupts ADXL Timer1 interrupts
 * \dontinclude adxl.c
 * \skip //ADXL TIMER1 interrupts begin
 * \until //ADXL TIMER1 interrupts end
 *
 * \subsection ADXL_MainPage_Timer2_Interrupts RTC and Timers Timer2 interrupts
 * \dontinclude timers.c
 * \skip //TIMER2 Interrupts begin
 * \until //TIMER2 Interrupts end
 * 
 * \subsection ADXL_MainPage_INT1_Interrupts Keyboard wakeup / RS232 INT1 intterupt
 * 
 * \subsection ADXL_MainPage_MainLoop Main Loop
 * Wraps everything together:
 * - Blinking of various LEDs to indicate that the AVR is alive
 * - The LCD charge pump which generates
 *   the negative voltage needed by the LCD
 * - Periodically monitoring the LCD voltage and
 *   do a PI regulation of the LCD bias voltage
 * - Update display with ADXL measurements
 * - Clock allocation (if CPU clock is needed for some perpripheral;
 *   ADXL, UART, Timer0
 * - Going to Power Save / Staying in idle
 * - Glue between printf and LCD, glue between dbg_printf and UART
 * .
 * \dontinclude accelR8.c
 * \skip //Main program loop begin
 * \until //Main program loop end
 * 
 * \section ADXL_MainPage_Download Download
 * 
 * <a href="./bullit_accelerometer.tar.gz">Bullit Accelerometer</a>
 */

/**
 * @page ADXL_Makefile_Page makefile
 * \verbinclude makefile
 */

/**
 * @page ADXL_References More Info
 * \subsection ADXL_References_Atmel Atmel
 * http://www.atmel.com
 *
 * \subsection ADXL_References_AnalogMems Analog Devices
 * http://www.analog.com 
 * http://www.analog.com/technology/mems/accelerometers/index.html
 * http://products.analog.com/products/info.asp?product=ADXL202
 *
 * \subsection ADXL_References_ChargePumps Charge Pumps
 * http://www.solarbotics.net/library/circuits/misc_pump.html
 *
 * \subsection ADXL_References_I2C I2C
 * http://www.philips.com/i2c
 *
 * \subsection ADXL_References_PhilipsDisplay LCD Display Driver Circuit
 *
 * \subsection ADXL_References_Doxygen Doxygen
 * http://www.doxygen.org
 *
 * \subsection  ADXL_References_Cygwin Cygwin
 * http://www.cygwin.org
 *
 * \subsection  ADXL_References_Tools Tools
 * http://www.avrfreaks.com
 * 
 * \subsection  ADXL_References_FIRFilters FIR Filters
 * http://dspguru.com/info/faqs/fir/basics.htm 
 */



/**
 * @page ADXL_Hardware_Page ADXL Hardware
 * The sensor is the \ref ADXL_References_AnalogMems "Analog Devices ADXL202". 
 *
 * CPU is an \ref ADXL_References_Atmel "ATMEL AT90S8535" running at 4Mhz.
 * A 32768 Hz crystal is connected to the real time clock timer inputs (TIMER2) and is the
 * time base for the accelerometer. This allows a ceramic resonator to be used
 * instead of a crystal if desired. A ceramic resonator should provide even lower power
 * consumption. Power to all subcomponents like display and accelerometer senser are
 * controlled by the CPU so that they can be turned off to save power.
 *
 * The ADXL is connected to the Input Capture Pin (ICP) and the analog comparator. Rset of the
 * ADXL that controls the period time is set so that there is approximatly 20000 ticks to one period.
 * 
 * The LCD display have 3 rows and is connected to the CPU using two data pins. 
 * The LCD uses \ref ADXL_References_I2C "I2C protocol"
 * and uses a variant of \ref ADXL_References_PhilipsDisplay "this driver circuit".
 * A negative voltage \ref ADXL_References_ChargePumps "charge pump" is driven by the CPU to generate
 * the bias voltage needed by the display. By measuring the bias voltage voltage with the ADC the CPU can
 * regulate the negative voltage. A \ref ADXL_Hardware_VLCDBias "voltage divider" between
 * the negative lcd bias and the battery makes the voltage in a range to be measured.
 *
 * Vref to the ADC is connected directly to the battery. One of the ADC inputs is connected to a reference
 * voltage from a zener diode. The zener is driven by one of the pins from the CPU.
 * This minimises the current consumption by just turning on the reference voltage
 * when Vbatt need to be calculated from measurements of the zener voltage.
 * 
 * RS232 line driver. A simple transistor inverts the output from the Tx line
 * so that it can be connected to an PC RS232 port. Two buttons are connected. Through two diodes they each
 * can pull down INT1 to wake the CPU if necessary. Three LED provide debug status, 
 * TIMER2 ticks, Idle/PowerSave and On/Off. Audio feedback is provided through a small buzzer.
 *
 */


/** @page ADXL_UserManual_Features Features
 * The accelerometer can measure acceleration and time. Using
 * the time and the acceleration it can calculate and display:
 * - Time (s)
 * - Average acceleration (m/s<SUP>2</SUP>)
 * - Instantanous acceleration (min/max) (m/s<SUP>2</SUP>)
 * - Speed (m/s & km/h)
 * - Distance (m & km)
 * .
 * The mass of the object can also be input, the accelerometer
 * can then also display:
 * - Instantenous Force (N)
 * - Average Force (N)
 * - Instantenous Power (kW and HP)
 * .
 */


/**
 * @page ADXL_UserManual_Page User Manual
 * \section ADXL_UserManual_Intro Introduction
 * @copydoc ADXL_UserManual_Features
 * \section ADXL_UserManual_Calibration Calibration 
 * Before the accelerometer can be used for measurements
 * it has to be calibrated and adjusted. 
 * Adjustment is done by keeping the accelerometer
 * vertical and rotating the accelerometer slowly 360 degreedes so that
 * each axis is subjected to +-1g.
 * 
 * For extra adjustments the local g-force can be input in terms of m/s<SUP>2</SUP>.
 * Default is 9.82m/s<SUP>2</SUP>. As the precision from the ADXL is limited
 * no adjustment of the g-force value is usually necessary.
 *
 * Calibrated values will be stored in the EEPROM.
 *
 * \section ADXL_UserManual_Usage Usage
 * The accelerometer can be set to measure in a varity of different modes.
 *
 * \subsection ADXL_UserManual_Usage_Instantenious Instantenous
 * Current acceleration, speed and distance is displayed.
 * RESET sets time and distance to zero.
 *
 * \subsection ADXL_UserManual_Usage_0_100 0-100
 * Measures the time from 0 km/h to 100 km/h. RESET resets time and distance
 * and enables the measurement. Time is started when speed is above 0.5m/s and
 * stopped when 100km/h is reached.
 *
 * \subsection ADXL_UserManual_Usage_0_100_0 0-100-0
 * \subsection ADXL_UserManual_Usage_QuarterMile Quarter Mile
 *
 */



