/*********************************************************************
 *   _      _        _     _                             _           *
 *  | |__  (_) _ __ | | __| |  ___  _ __    _ __    ___ | |_         *
 *  | '_ \ | || '__|| |/ /| | / _ \| '__|  | '_ \  / _ \| __|        *
 *  | |_) || || |   |   < | ||  __/| |   _ | | | ||  __/| |_         *
 *  |_.__/ |_||_|   |_|\_\|_| \___||_|  (_)|_| |_| \___| \__|        *
 *                                                                   *
 *********************************************************************
 * Copyright 2002                                                    *
 * Jrgen Birkler (birkler@yahoo.com)                                *
 *********************************************************************/

/**
 * @file
 * ADXL Analog Devices Accelormeter Utilities
 * @ingroup ATMEL_AVR_Embedded_ADXL
 * @author Jrgen Birkler (birkler@yahoo.com) 
 */

#include "hardware.h"
#include "avr/signal.h"
#include "avr/interrupt.h"
#include "debug.h"
#include "adxl.h"
#include "fir_filter.h"

//typedef uint32_t uS_t;


//Analog Comparator
////////////////////////////
#define ACOMP_ENABLE() cbi(ACSR,ACD)
#define ACOMP_DISABLE() sbi(ACSR,ACD)


// Select Analog comparator or ICP
////////////////////////////////////////

///Select x-axis (ICP)
#define ADXL_SELECT_X_AXIS() cbi(ACSR,ACIC)
///Select y-axis (Analog Comparator)
#define ADXL_SELECT_Y_AXIS() sbi(ACSR,ACIC)
///@returns true if x-axis selected, false if y-axis is selected
#define ADXL_IS_X_AXIS() bit_is_clear(ACSR,ACIC)



//Timer 1
////////////////////////////////
#define TIMER1_INPUT_CAPTURE_RISING_EDGE() sbi(TCCR1B,ICES1)
#define TIMER1_INPUT_CAPTURE_FALLING_EDGE() cbi(TCCR1B,ICES1)

#define TIMER1_INPUT_CAPTURE_IS_RISING_EDGE() bit_is_set(TCCR1B,ICES1)

#define TIMER1_ENABLE_INPUT_CAPTURE_INTERRUPT() {sbi(TIMSK,ICF1);}
#define TIMER1_ENABLE_OUTPUT_COMPARE_A_INTERRUPT() {sbi(TIMSK,OCIE1A);}

#define TIMER1_DISABLE_INPUT_CAPTURE_INTERRUPT() {cbi(TIMSK,ICF1);}
#define TIMER1_DISABLE_OUTPUT_COMPARE_A_INTERRUPT() {cbi(TIMSK,OCIE1A);}

#define TIMER1_SET_OUTPUT_COMPARE_A(_value)  OCR1A = (_value)

#define TIMER1_INIT() \
{ \
	/*No PWM, no output compare*/ \
	TCCR1A = 0; \
	/*Enable ICP, enable noise canceler*/ \
	TCCR1B = BV(ICNC1); \
}

#define TIMER1_CLEAR_COUNTER() outw(0,TCNT1)
#define TIMER1_READ_COUNTER()   inw(TCNT1)
#define TIMER1_READ_INPUT_CAPTURE()   inw(ICR1)
#define TIMER1_ENABLE_OVERFLOW_INTERRUPT() 
#define TIMER1_CLK CLOCK_FREQ
#define TIMER1_TICKS_TO_uS(ticks) (ticks * 1000000L / TIMER1_CLK)

#define TIMER1_GET_PRESCALER() (TCCR1B & (BV(CS12) | BV(CS11) | BV(CS10))) 

#define TIMER1_SET_PRESCALER(prescaler) TCCR1B = (TCCR1B & (BV(CS12) | BV(CS11) | BV(CS10))) | prescaler

#define TIMER1_PRESCALER_CLOCK_OFF (0)
#define TIMER1_PRESCALER_X1    (BV(CS10))
#define TIMER1_PRESCALER_X8    (BV(CS11))
#define TIMER1_PRESCALER_X64   (BV(CS11)|BV(CS10))
#define TIMER1_PRESCALER_X256  (BV(CS12))
#define TIMER1_PRESCALER_X1024 (BV(CS12)|BV(CS10))

#if (ADXL_AUTOMATIC_PRESCALER_ENABLED == 1)

static unsigned char Timer1PrescalerFactor;

static void Timer1_Prescaler_Increase(void);
static unsigned char Timer1_Prescaler_Decrease(uint16_t ticks);
inline static uint8_t Timer1_GetPrescalerFactor(void);
inline static void Timer1_SetPrescalerFactor(uint8_t value);

#endif //(ADXL_AUTOMATIC_PRESCALER_ENABLED == 1)



//Variables
//////////////////////

///Size of floating average
#define ADXL_PERIOD_FILTER_LOG2SIZE (5)

///Number of measurements before values are stable
#define ADXL_STABLE_MEASUREMENT_COUNT (1<<ADXL_PERIOD_FILTER_LOG2SIZE)


//Define fir filter for period
FIR_FILTER_DEFINE(periodFilter, ADXL_PERIOD_FILTER_LOG2SIZE);

//X-Axis fir filter 
FIR_FILTER_DEFINE(AxisXFilter, ADXL_PERIOD_FILTER_LOG2SIZE);

//Y_axis fir filter
FIR_FILTER_DEFINE(AxisYFilter, ADXL_PERIOD_FILTER_LOG2SIZE);

///Current AxisX value
static volatile ADXL_TimerTicks_t AxisX;

///Current AxisY value
static volatile ADXL_TimerTicks_t AxisY;

///0 if measurements are valid
static volatile unsigned char validMeasurements;  

///Set calibration info
static ADXL_CalibrationInfo_t CalibrationSet;

///Current measure calibration values
static ADXL_CalibrationInfo_t CalibrationMeasured;








///Max period ticks before overrun
#define ADXL_TIMERTICKS_OVERRUN 0x7FFF



///Saved Ta or Tc
static uint16_t TaTc;


//ADXL TIMER1 interrupts begin
INTERRUPT(SIG_INPUT_CAPTURE1)
{
  //Get the captured counter  
	uint16_t cnt = TIMER1_READ_INPUT_CAPTURE();

	//Normalize to ticks after TaTc, use the fact that it is 
	//16 bit timer and 16bit arithmetics is used
  cnt -= TaTc;

	//Check prescaler, if cnt < 65536 / 4 then we can use a lower
	//prescaler value
	#if (ADXL_AUTOMATIC_PRESCALER_ENABLED == 1)
	if (!Timer1_Prescaler_Decrease(cnt))
	#endif //(ADXL_AUTOMATIC_PRESCALER_ENABLED == 1)
	{
		if (validMeasurements > 0)
    {
			validMeasurements--;
    }
    else
    {
        if (AxisX > CalibrationMeasured.TxMax)
          CalibrationMeasured.TxMax = AxisX;
        if (AxisX < CalibrationMeasured.TxMin)
          CalibrationMeasured.TxMin = AxisX;
        if (AxisY > CalibrationMeasured.TyMax)
          CalibrationMeasured.TyMax = AxisY;
        if (AxisY < CalibrationMeasured.TyMin)
          CalibrationMeasured.TyMin = AxisY;
    }

		//See algorithm description in .h file.
		if (TIMER1_INPUT_CAPTURE_IS_RISING_EDGE())
		{
			//(a) (c)
      TaTc += cnt;

			CalibrationMeasured.Tp = 
        FIR_Filter_BoxCar_Input(&periodFilter.filter, (FIR_FilterSample_t)cnt);

			//Enable overflow detection
      TIMER1_SET_OUTPUT_COMPARE_A(TaTc + ADXL_TIMERTICKS_OVERRUN);
			TIMER1_INPUT_CAPTURE_FALLING_EDGE();
		}
		else {
      // (b) (d)
      if (ADXL_IS_X_AXIS())
      {
			  //(b)
        AxisX = FIR_Filter_BoxCar_Input(&AxisXFilter.filter, (FIR_FilterSample_t)cnt);

        ADXL_SELECT_Y_AXIS();
      }
      else {
        // (d)
        AxisY = FIR_Filter_BoxCar_Input(&AxisYFilter.filter, (FIR_FilterSample_t)cnt);

        ADXL_SELECT_X_AXIS();
      }
			TIMER1_INPUT_CAPTURE_RISING_EDGE();
		}
	}
}




INTERRUPT(SIG_OUTPUT_COMPARE1A)
{
	//Timer 1 overflowed, increase prescaler and reset the number of valid validMeasurements
	//uS_counter = 0;

#if (ADXL_AUTOMATIC_PRESCALER_ENABLED == 1)
	Timer1_Prescaler_Increase();
#endif //(ADXL_AUTOMATIC_PRESCALER_ENABLED == 1)
	RED_LED_CHANGE();

	validMeasurements = ADXL_STABLE_MEASUREMENT_COUNT;
}


//ADXL TIMER1 interrupts end





void ADXL_PowerOn(void)
{
	ADXL_POWER_ON();
  ACOMP_ENABLE();
  ADXL_SELECT_Y_AXIS();

#if (ADXL_AUTOMATIC_PRESCALER_ENABLED == 1)
	Timer1PrescalerFactor = 0; 
#endif //(ADXL_AUTOMATIC_PRESCALER_ENABLED == 1)
	TIMER1_SET_PRESCALER(TIMER1_PRESCALER_X1); 
  
	FIR_Filter_Init(&periodFilter.filter,ADXL_PERIOD_FILTER_LOG2SIZE);
	FIR_Filter_Init(&AxisXFilter.filter,ADXL_PERIOD_FILTER_LOG2SIZE);
	FIR_Filter_Init(&AxisYFilter.filter,ADXL_PERIOD_FILTER_LOG2SIZE);
	validMeasurements = ADXL_STABLE_MEASUREMENT_COUNT;

  TIMER1_INPUT_CAPTURE_RISING_EDGE();
  TIMER1_ENABLE_INPUT_CAPTURE_INTERRUPT();
  TIMER1_DISABLE_OUTPUT_COMPARE_A_INTERRUPT();
}


void ADXL_PowerOff(void)
{
	//Disable interrupts
  TIMER1_DISABLE_INPUT_CAPTURE_INTERRUPT();
  TIMER1_DISABLE_OUTPUT_COMPARE_A_INTERRUPT();

	TIMER1_SET_PRESCALER(TIMER1_PRESCALER_CLOCK_OFF);

  ACOMP_DISABLE();

  //Power down accelerometer
	ADXL_POWER_OFF();
}




void ADXL_SetCalibrationInfo(const ADXL_CalibrationInfo_t* const info_p)
{
	CalibrationSet = *info_p;
}

void ADXL_ResetMeasuredCalibrationInfo(void)
{
	CalibrationMeasured.Tp = ADXL_TIMERTICKS_MAX;
	CalibrationMeasured.TxMin = ADXL_TIMERTICKS_MAX;
	CalibrationMeasured.TxMax = 0;
	CalibrationMeasured.TyMin = ADXL_TIMERTICKS_MAX;
	CalibrationMeasured.TyMax = 0;
}


const ADXL_CalibrationInfo_t* ADXL_GetMeasuredCalibrationInfo(void)
{
	return &CalibrationMeasured;
}





unsigned char ADXL_MeasurementAvailable(void)
{
  return (validMeasurements==0);
}


static inline ADXL_TimerTicks_t ADXL_GetPeriod(void)
{
  return CalibrationMeasured.Tp;
}


///Large value type used for calculations
typedef FIR_FilterAcculatedSum_t ADXL_long_mdG_t;



/*
 *              2 * Tpulse * Tpcal - (Tmin+Tmax) * Tperiod 
 * A(g) = 1g * --------------------------------------------
 *                     2 * (Tmax - Tmin) * Tperiod
 */
ADXL_g_t ADXL_GetX(void)
{
  ADXL_TimerTicks_t Tp = ADXL_GetPeriod();
	ADXL_long_mdG_t meas;
  ADXL_long_mdG_t calib;

  cli();
	meas = 2 * AxisX;
  sei();
  meas *= CalibrationSet.Tp;

  calib = (CalibrationSet.TxMax + CalibrationSet.TxMin);
  calib *= Tp;

  meas -= calib;

  meas /= CalibrationSet.TxMax - CalibrationSet.TxMin;

  meas *= 10000;

  meas /= Tp;

	return (ADXL_g_t)meas;
}

ADXL_g_t ADXL_GetY(void)
{
  ADXL_TimerTicks_t Tp = ADXL_GetPeriod();
	ADXL_long_mdG_t meas;
  ADXL_long_mdG_t calib;

  cli();
	meas = 2 * AxisY;
  sei();
  meas *= CalibrationSet.Tp;

  calib = CalibrationSet.TyMax + CalibrationSet.TyMin;
  calib *= Tp;

  meas -= calib;

  meas /= CalibrationSet.TyMax - CalibrationSet.TyMin;

  meas *= 10000;

  meas /= Tp;

	return (ADXL_g_t)meas;
}



//Timer 1 helpers
//////////////////////////////////////////////


#if (ADXL_AUTOMATIC_PRESCALER_ENABLED == 1)

static void Timer1_Prescaler_Increase(void)
{
	DEBUG_PRINT_INFO_("<ps+>");

	switch (TIMER1_GET_PRESCALER())
	{
		case TIMER1_PRESCALER_X1:
			TIMER1_SET_PRESCALER(TIMER1_PRESCALER_X8);
			Timer1PrescalerFactor = 3;
			break;
		case TIMER1_PRESCALER_X8:
			TIMER1_SET_PRESCALER(TIMER1_PRESCALER_X64);
			Timer1PrescalerFactor = 6;
			break;
		case TIMER1_PRESCALER_X64:
			TIMER1_SET_PRESCALER(TIMER1_PRESCALER_X256);
			Timer1PrescalerFactor = 8;
			break;
		case TIMER1_PRESCALER_X256:
			TIMER1_SET_PRESCALER(TIMER1_PRESCALER_X1024);
			Timer1PrescalerFactor = 10;
			break;
		default:
			break;
	}
}

static unsigned char Timer1_Prescaler_Decrease(uint16_t ticks)
{
	ticks=ticks;
	return false;

	if (ticks < (0x0FFF) )
	{
		DEBUG_PRINT_INFO_("<ps->");

		switch (TIMER1_GET_PRESCALER())
		{
			case TIMER1_PRESCALER_X1024:
				TIMER1_SET_PRESCALER(TIMER1_PRESCALER_X256);
				Timer1PrescalerFactor = 8;
				return true;
			case TIMER1_PRESCALER_X256:
				TIMER1_SET_PRESCALER(TIMER1_PRESCALER_X64);
				Timer1PrescalerFactor = 6;
				return true;
			case TIMER1_PRESCALER_X64:
				TIMER1_SET_PRESCALER(TIMER1_PRESCALER_X8);
				Timer1PrescalerFactor = 3;
				return true;
			case TIMER1_PRESCALER_X8:
				TIMER1_SET_PRESCALER(TIMER1_PRESCALER_X1);
				Timer1PrescalerFactor = 0;
				return true;
			default:
				return false;
		}
	}
	else {
		return false;
	} 
}


static uint8_t Timer1_GetPrescalerFactor(void)
{
	return Timer1PrescalerFactor;
}



#endif //(ADXL_AUTOMATIC_PRESCALER_ENABLED == 1)
