/*********************************************************************
 *   _            _  _  _   _      _                             _           *
 *  | |__   _  _ | || |(_) | |_ 
 *  | '_ \ | || || || || | | __|  *
 *  | |_) || || || || || | | |_     *
 *  |_.__/  \___||_||_||_|  \__|        *
 *                                                                   *
 *********************************************************************
 * Copyright 2002                                                    *
 * Jrgen Birkler (birkler@yahoo.com)                                *
 *********************************************************************/

/** 
 * @file
 * ADXL Analog Devices Acceloremeter Utilities
 * @ingroup ADXL_
 * @author Jrgen Birkler (birkler@yahoo.com) 
 * @copydoc ADXL_TheoryOfOperation
 */
         


/** @page ADXL_TheoryOfOperation ADXL - Theory of Operation
 *
 * \section ADXL_TheoryOfOperation1 Theory of operation
 * \subsection ADXL_TheoryOfOperation1 PWM Signal
 * The ADXL outputs the acceleration as a PWM signal. The ratio between the
 * pulse length and the period length is the measured acceleration. 
 *
 * A(g) = (Tpulse / Tperiod - 50%) / 12.5%/g
 *
 * 50% means 0g, 75% means +2g and 25% means -2g.
 *
 * \code
 *                    Tperiod
 *               |---------------|
 *          (a)     (b)                     (a)     (b)
 *            _______         _______         _______         _______
 * X-AXIS ___/       \_______/       \_______/       \_______/       \_______
 *             _____           _____           _____           _____
 * Y-AXIS ____/     \_________/     \_________/     \_________/     \________
 *                           (c)   (d)                       (c)   (d)
 *
 * \endcode
 * a,b,c,d are the different events recorded by the input capture routine
 *
 * \subsection ADXL_TheoryOfOperation2 TIMER1 Input Capture
 * The ICP (Input Capture Pin) and Acomp (Analog Comparator) is used
 * sample the x-axis and the y-axis alternativly. One axis is connected
 * to the ICP pin and one axis is connected to the positive input of the
 * analog comparator. The negative input of the comparator is biased with half
 * the supply to the ADXL using two 47K resistors.
 * Period time (Tp) is the average of \e "Time from rising edge of x-axis to rising edge of y-axis"
 * and \e "Time from rising edge y-axis to rising edge of x-axis"
 *
 * This algorithm is somewhat different than the one described in the
 * application note. I think this algorithm is easier to understand and
 * requires less computations. 
 *
 * The two different axis can be selected by changing the Analog Comparator Input Capture
 * Enable bit in the Analog Comparator Control Status Register.
 *
 * Ta = Counter sampled at (a)
 *
 * Tc = Counter sampled at (c)
 *
 * Tac = Time from rising X-axis edge to rising Y-axis edge
 *
 * Tca = Time from rising Y-axis edge to rising X-axis edge
 *
 * Tp = = Tperiod = Period Time = (Tac + Tca) / 2
 *
 * Tx = Tab = length of X-axis pulse
 *
 * Ty = Tcd = length of Y-axis pulse
 *
 * - (a) Rising edge of X-Axis 
 *   - Tca = CNT - Tc
 *   - Save Ta
 *   - Enable output compare (Ta) overflow
 *   - Tp = (Tac + Tca) / 2
 *   - Enable X-axis falling edge
 * - (b) Falling edge of X-Axis 
 *   - Tx = CNT - Ta
 *   - Switch to Y-axis
 *   - Enable Y-axis rising edge
 * - (c) Rising edge of Y-Axis  
 *   - Tac = CNT - Ta
 *   - Save Tc
 *   - Enable output compare (Tc) overflow
 *   - Tp = (Tca + Tac) / 2
 *   - Enable Y-axis falling edge
 * - (d) Falling edge of Y-Axis 
 *   - Ty = CNT - Tc
 *   - Switch to X-axis
 *   - Enable X-axis rising edge
 * .
 * Ta and Tc are stored in one \ref TaTc "shared variable". The counter is never reset and is free running,
 * it restarts at zero after 65355. 16 bit arethmetics is used to get correct number of ticks. 
 * To detect timer overflows (when period time is to long) 
 * output compare is enabled at (a) and (c) .
 * If the interrupt triggers it means that the period time is to long and
 * the prescaler or the Rset for the ADXL202 need to be adjusted.
 *
 * \subsection ADXL_TheoryOfOperation3 Calibration
 * By rotating the accelerometer 360 or more degrees both axis are subjected to +-1g. 
 * Calibration info is recorded during this rotation. #ADXL_CalibrationInfo_t.
 * See also \ref ADXL_UserManual_Page.
 *
 * During the rotation the minimum (Tmin)
 * and maximum (Tmax) pulse length for each axis are saved along
 * with the period time (Tpcal). 
 *
 * \subsection ADXL_TheoryOfOperation4 Calculations
 * The zero g offset ratio (nomally 50%) is:
 *
 * Z = (Tmin+Tmax) / (2*Tpcal)
 *
 * The scale (nomally 12.5%/g) is:
 *
 * S = ((Tmax - Tmin) / Tpcal) / 2.0g = (Tmax - Tmin) / (2.0g * Tpcal)
 *
 * Expressed using zero g offset (Z) and scale (S) the acceleration is:
 * 
 * A(g) = ((Tpulse / Tperiod) - Z) / S
 *
 * \code
 *         Tpulse / Tperiod - (Tmin+Tmax) / (2 * Tpcal)
 * A(g) = ----------------------------------------------
 *              (Tmax - Tmin) / (2.0g * Tpcal)
 * \endcode
 * Move (2.0g * Tpcal) division:
 * \code
 *         Tpulse * (Tpcal * 2.0g) / Tperiod - (Tmin+Tmax) * (Tpcal * 2.0g) / (2*Tpcal)
 * A(g) = --------------------------------------------------------------------------
 *                                (Tmax - Tmin)  
 * \endcode
 * Multiply nomand denom by Tperiod, break out 1g and skip 2*Tpcal from subtraction element:
 * \code
 *                2 * Tpulse * Tpcal - (Tmin+Tmax) * Tperiod 
 * A(g) = 1.0g * --------------------------------------------
 *                       (Tmax - Tmin) * Tperiod
 * \endcode
 */
                                                         

/**
 * @defgroup ADXL_ Analog Devices Accelerometer Utilities
 * Functions for measuring acceleration using Analog Devices
 * ADXL202.
 *
 * @ingroup ATMEL_AVR_Embedded
 * 
 * @copydoc ADXL_TheoryOfOperation
 */





/** 
 * \addtogroup ADXL_
 * \@{
 */


#ifndef __ADXL_MEASURE_H__
#define __ADXL_MEASURE_H__

#include <inttypes.h>


///Timer ticks counter from ADXL202
typedef uint16_t ADXL_TimerTicks_t;


///Max ticks
#define ADXL_TIMERTICKS_MAX 0xFFFF



///deci milli g. +10000dmg = +1g
typedef int16_t ADXL_g_t;


/**
 * Calibration value as per being subjected to +-1g
 */
typedef struct {
	ADXL_TimerTicks_t Tp;     ///<Period ticks during calibration
	ADXL_TimerTicks_t TxMin;  ///<X-axis -1g ticks
	ADXL_TimerTicks_t TxMax;  ///<Y-axis +1g ticks
	ADXL_TimerTicks_t TyMin;  ///<X-axis -1g ticks
	ADXL_TimerTicks_t TyMax;  ///<Y-axis +1g ticks
} ADXL_CalibrationInfo_t;


///Uncalibrated EEPROM #ADXL_CalibrationInfo_t settings
#define ADXL_CALIBRATION_INITIAL_VALUE {200,75,125,75,125}

/**
 * Set new calibration info
 * Use this function to set calibration info read from EEPROM
 */
void ADXL_SetCalibrationInfo(const ADXL_CalibrationInfo_t* const info_p);



/**
 * Resets calibration info.
 *
 * Call this function before user start 360 degree +-1g calibration.
 * Use #ADXL_GetMeasuredCalibrationInfo to get calibration info when
 * calibration is done. This resets the measured calibration info.
 * When calibration is done #ADXL_SetCalibrationInfo must be called
 * for the new calibration to take affect. For example:
 * \code
 * ADXL_SetCalibrationInfo(ADXL_GetMeasuredCalibrationInfo());
 * \endcode
 * 
 */
void ADXL_ResetMeasuredCalibrationInfo(void);


/**
 * Get measure calibration info from ADXL library.
 * Call this when user has subjected the accelerotmeter to +-1g on both
 * axis.
 * #ADXL_ResetMeasuredCalibrationInfo must be called before calibration
 * is started.
 * @returns The measured calibration info. Store this to NV memory (EEPROM).
 *
 */
const ADXL_CalibrationInfo_t* ADXL_GetMeasuredCalibrationInfo(void);




/**
 * Start measuring acceleration
 * @see ADXL_MeasurementAvailable
 */
void ADXL_PowerOn(void);


/**
 * Returns != 0 if measurement is available
 */
unsigned char ADXL_MeasurementAvailable(void);


/**
 * Get X-Axis value
 * @returns X-axis g-force in deci-milli-g. +1g is output as +10000.
 * @note ADXL must have correct calibration info
 */
ADXL_g_t ADXL_GetX(void);


/**
 * Get Y-axis value
 * @returns X-axis g-force in deci-milli-g. +1g is output as +10000.
 * @note ADXL must have correct calibration info
 */
ADXL_g_t ADXL_GetY(void);



ADXL_TimerTicks_t ADXL_GetAverage(void);

/**
 * Stop measure and power down all resources
 * (ADXL, Timer1, Analog Compartor)
 */
void ADXL_PowerOff(void);


#endif //__ADXL_MEASURE_H__

