//*****************************
//* Hardware definitions
//*****************************
//* Jrgen Birkler (birkler@yahoo.com)
//* http://birkler.net
//* Copyright 2002
//*****************************

/**
 * @file
 * Hardware configuration
 *
 * @author Jrgen Birkler (birkler@yahoo.com)
 * @copydoc ADXL_HardwareAssignments
 */

/** @page ADXL_HardwareAssignments AT90S8535 hardware assignments
 * \code
 * PD7 [out] BLUE LED (1s alive blink)
 * PD6 [in]  ICP; ADXL x-axis (8535:pin20)
 * PD5 
 * PD4 [out] BUZZER_OUT
 * PD3 [in]  INT1. _WAKEUP  (8535:pin17)
 * PD2       INT0.  (8535:pin16)
 * PD1 [out] UART TXD (8535:pin15)
 * PD0 [in]  UART RXD (8535:pin14)
 *
 * PC7 [in]  32678 Hz Crystal
 * PC6 [out] 32678 Hz Crystal
 * PC5 [i/o] I2C Data (8535:pin27) 
 * PC4 [i/o] I2C Clk (8535:pin26)
 * PC3 [out] LCD_PWR  (8535:pin25)
 * PC2 [out] LCD_CHARGE_PUMP2  (8535:pin24)
 * PC1 [out] LCD_CHARGE_PUMP1  (8535:pin23)
 * PC0 [out] LCD_CHARGE_PUMP0  (8535:pin22)
 *
 * PB7 [in]  SCK   (8535:pin8)
 * PB6 [out]  MISO   (8535:pin7)
 * PB5 [in] MOSI   (8535:pin6)
 * PB4       _SS
 * PB3 [in]  AIN1;(ADXL Power / 2) (8535:pin4)
 * PB2 [in]  AIN0; ADXL y-axis (8535:pin3)
 * PB1 [out] ADXL Power (8535:pin2)
 * PB0 [out] RED LED (on if CPU clock is on, off is CPU is in power save)
 *
 * PA7
 * PA6   (8535:pin34)
 * PA5
 * PA4
 * PA3
 * PA2 [out] Vref supply  (8535:pin38)
 * PA1 [ain] VLCD Meas    (8535:pin39)
 * PA0 [ain] Vcc measure  (8535:pin40)
 * \endcode
 */

//#define DEBUG_DEBUGGER_





#ifndef __HARDWARE_H__
#define __HARDWARE_H__

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/signal.h>
#include <avr/sleep.h>


//CPU
///////////////////////

#ifdef DEBUG_TARGET_
#define CLOCK_FREQ (40000L)
#else
#define CLOCK_FREQ (4000000L)
#endif //CLOCK_FREQ

///Wait for interupt (oscillator and core is running)
#define CPU_WAIT() {set_sleep_mode(SLEEP_MODE_IDLE);sleep_mode();}


///Go to powersave. Oscillator and core is stopped. Only Async Timer 2 and Int0/Int1 can wake CPU.
#define CPU_POWER_SAVE() {set_sleep_mode(SLEEP_MODE_PWR_SAVE);sleep_mode();}

///Go to idle (oscillator is running, core is sleeping)
#define CPU_IDLE() {set_sleep_mode(SLEEP_MODE_IDLE);sleep_mode();}


///Init all port to input with enabled pull up to avoid current draw at power down
#define CPU_PORTS_INIT() { \
	DDRA=0;DDRB=0;DDRC=0;DDRD=0; \
	PORTA=~0;PORTB=~0;PORTC=~0;PORTD=~0; \
}

#define FOREVER for(;;)



///Init CPU
#define CPU_INIT() { \
	CPU_PORTS_INIT() \
}


//I2C
///////////////////////////

#define I2C_RELEASE_SDA() {sbi(PORTC,PC5);cbi(DDRC,DDC5);}
#define I2C_RELEASE_SCL() {sbi(PORTC,PC4);cbi(DDRC,DDC4);}
#define I2C_DRIVE_SDA()   {sbi(DDRC,DDC5);cbi(PORTC,PC5);}
#define I2C_DRIVE_SCL()   {sbi(DDRC,DDC4);cbi(PORTC,PC4);}
#define I2C_IS_SDA_DRIVEN()    (bit_is_clear(PINC,PC5))
#define I2C_IS_SCL_RELEASED()  (bit_is_set(PINC,PC4))


#define INT1_ENABLE_INTERRUPT_RISING()  {sbi(MCUCR,ISC11);sbi(MCUCR,ISC10);sbi(GIMSK,INT1);}
#define INT1_ENABLE_INTERRUPT_FALLING() {sbi(MCUCR,ISC11);cbi(MCUCR,ISC10);sbi(GIMSK,INT1);}
#define INT1_IS_RISING_EDGE()           bit_is_set(MCUCR,ISC10)
#define INT1_DISABLE_INTERRUPT()        {cbi(GIMSK,INT1);}

//WAKEUP Signal (INT1)
///////////////////////////////////
#define WAKEUP_INIT() {cbi(DDRD,DDD3);sbi(PORTD,PD3);INT1_ENABLE_INTERRUPT_FALLING();}


//Keyboard
///////////////////////////////////

#define BUTTONS_INIT() {cbi(DDRA,DDA7);cbi(DDRA,DDA6);sbi(PORTA,PA7);sbi(PORTA,PA6);}
#define BUTTON_RESET_DOWN() bit_is_clear(PINA,PA7)
#define BUTTON_MODE_DOWN() bit_is_clear(PINA,PA6)


/*

#define KEYBOARD_KEYS 3
#define KEY_LEFT     1
#define KEY_RIGHT     2
#define KEY_SELECT    3
#define KEYBOARD_KEY_IS_DOWN(i) (!(PINC & (0x04<<(i))))
#define KEYBOARD_INIT()  { DDRC &= ~0x38; PORTC |= 0x38;}
*/



//Analogue inputs
////////////////////////////////////
//All in mV

#define ADC_REF_VOLTAGE (4500)

#define ADC_TIMER_COUNT (CLOCK_FREQ/100)

#define ADC_INIT()              {DDRA = 0;PORTA=0;}
//Prescaler set to /2
#define ADC_ENABLE()            ADCSR = BV(ADEN) /*| BV(ADPS2) | BV(ADPS0)*/
#define ADC_SETUP_MUX(InputMux) ADMUX=((InputMux)&0x07)
#define ADC_GET_MUX()           (ADMUX)
#define ADC_ENABLE_INTERRUPT()  sbi(ADCSR,ADIE)
#define ADC_DISABLE_INTERRUPT() cbi(ADCSR,ADIE)
#define ADC_START()             sbi(ADCSR,ADSC)
#define ADC_DONE()              is_bit_clear(ADCSR,ADSC)
#define ADC_DISABLE()           outb(ADCSR,0)
#define ADC_GET_VALUE()         inw(ADCL)

/*
#define    ADEN     7
#define    ADSC     6
#define    ADFR     5
#define    ADIF     4
#define    ADIE     3
#define    ADPS2    2
#define    ADPS1    1
#define    ADPS0    0  
 */


#define ADC_MUX_VLCD (1)


#define ADC_SCALE_FACTOR (7)
#define ADC_VBATT (4600)
#define ADC_VREF ADC_VBATT

#define ADC_MEASUREMENT_TO_VOLTAGE(measure) ((measure)*((ADC_VREF)>>(ADC_SCALE_FACTOR)) / (1024>>(ADC_SCALE_FACTOR)))

//Temperature sensor: 10mV per centigrade. 0 centigrade = 2.73V
#define ADC_VLCD(measused_voltage) ((measused_voltage) - ADC_REF_VOLTAGE) * 

#define ADC_KELVIN_PER_VOLT (100.0)
#define ADC_CELCIUS_AT_0_KELVIN (-273.0)
#define ADC_VOLT_TO_KELVIN(voltage) ((voltage)*ADC_KELVIN_PER_VOLT)
#define ADC_KELVIN_TO_CENTIGRADE(kelvin) ((kelvin) + ADC_CELCIUS_AT_0_KELVIN)
#define ADC_VOLT_TO_CENTIGRADE(voltage) ADC_KELVIN_TO_CENTIGRADE(ADC_VOLT_TO_KELVIN(voltage))


/**
 * \page ADXL_Hardware_VLCDBias
 * \code
 *                  +-----o Vbatt
 *                  |
 *               R1 #     -
 *                  #       } Vr1
 *                  |     +
 *                  +---------o Vmeas
 *                  |
 *               R2 #       +-
 *                  #       |
 *                  |      ---
 *         Vlcd o---+      GND
 *
 * 
 * ADC_LCD_V_MEAS = ADC_VBATT + ADC_LCD_V_R1
 * ADC_LCD_V_R1 = (ADC_LCD_V_LCD - ADC_VBATT) * (ADC_LCD_R1 / (ADC_LCD_R1 + ADC_LCD_R2))
 * =>
 * (ADC_LCD_R1 + ADC_LCD_R2) * (ADC_LCD_V_MEAS - ADC_VBATT) = ADC_LCD_R1 * (ADC_LCD_V_LCD - ADC_VREF)
 * =>
 * ADC_LCD_V_LCD = ((ADC_LCD_R1 + ADC_LCD_R2) * (ADC_LCD_V_MEAS - ADC_VBATT)) / ADC_LCD_R1 + ADC_VBATT
 * \endcode
 */

///@copydoc ADXL_Hardware_VLCDBias
#define ADC_LCD_V_LCD(_v_meas) (((ADC_LCD_R1 + ADC_LCD_R2) * ((_v_meas) - ADC_VBATT)) / ADC_LCD_R1 + ADC_VBATT)


#define ADC_LCD_R1 1 //680K
#define ADC_LCD_R2 2 //1.25M

///mV
typedef signed int Voltage_t;


//ADXL202
/////////////////////////

#define ADXL_INIT() {cbi(DDRB,DDB2);cbi(PORTB,PB2); cbi(DDRB,DDB3);cbi(PORTB,PB3); cbi(DDRD,DDD6);cbi(PORTD,PD6); sbi(PORTB,PB1);sbi(DDRB,DDB1);}
#define ADXL_POWER_ON() {cbi(PORTB,PB1);}
#define ADXL_POWER_OFF() {sbi(PORTB,PB1);}




//UART
//////////////////////////

/**
 * Waits for the transmitter to be ready then outputs
 * the character
 */
#define UART_WRITE_BYTE(_ch) {\
	loop_until_bit_is_set(USR,UDRE); \
	UDR = (_ch); }
	


#define UART_WAIT_BYTE() \
	loop_until_bit_is_set(USR,RXC) 

/**
 * Waits until a character becomes available
 * then sets
 */
#define UART_READ_BYTE() \
	UDR 


#define UART_BYTE_IS_AVAILABLE() bit_is_set(USR,RXC) 

#define UART_WAIT_TRANSMITTER_AVAILABLE() loop_until_bit_is_set(USR,UDRE);  
#define UART_TXC_INTERRUPT_ENABLE() sbi()
#define UART_TXC_INTERRUPT_DISABLE() sbi()


#define UART_GET_UBRR(baud) ((CLOCK_FREQ/((baud)*16L))-1)
#define UART_RX_INTERRUPT_ENABLE() sbi(UCR,RXCIE)
#define UART_TX_INTERRUPT_ENABLE() 

#define UART_INIT(baud) { \
	sbi(DDRD,DDD1); \
	sbi(PORTD,PD1); \
	UBRR = UART_GET_UBRR(baud); \
	UCR |= BV(RXEN) | BV(TXEN); \
  }
	

#define UART_POWER_SAVE_SLEEP()  { \
	volatile uint16_t i; \
	for (i=1255;i-->0;) {} \
	cbi(UCR,TXEN); \
}

//	UART_WRITE_BYTE('\r'); 
//	loop_until_bit_is_set(USR,TXC); 


#define UART_POWER_SAVE_WAKEUP() sbi(UCR,TXEN)



//LEDS
////////////////////////////////////////////
#define BLUE_LED_ON()        cbi(PORTD,PD7)
#define BLUE_LED_OFF()       sbi(PORTD,PD7)
#define BLUE_LED_CHANGE()    PORTD^=BV(PD7)
#define LEDS_INIT()          {sbi(DDRD,DDD7);sbi(DDRB,DDB0);}
#define RED_LED_ON()         cbi(PORTB,PB0)
#define RED_LED_OFF()        sbi(PORTB,PB0)
#define RED_LED_CHANGE()     PORTB^=BV(PB0)


//VLCD Charge Pump
/////////////////////////////////////////////
#define LCD_INIT() { \
	DDRC|=BV(DDC0) | BV(DDC1) | BV(DDC2) | BV(DDC3); \
	LCD_CHARGE_PUMP_OFF(); \
}

#define LCD_CHARGE_PUMP_TOGGLE() PORTC^=BV(PC0) | BV(PC1) | BV(PC2)
#define LCD_CHARGE_PUMP_OFF() PORTC = (PORTC & ~BV(PC1)) | BV(PC0) | BV(PC2)

#define LCD_POWER_ON() sbi(PORTC,PC3)

#define LCD_POWER_OFF() cbi(PORTC,PC3)


//RTC and LED PWM pump timer generator
////////////////////////////////////////////////

#define TIMER2_CLOCK_FREQ       (32768) //Hz
#define TIMER2_RTC_UPDATE_TIME  (1000) //ms

//Prescalers available are 1,8,32,64,128,256,1024


#define TIMER2_DEBUG_OUTPUT_INIT() //sbi(DDRB,DDB0)
#define TIMER2_DEBUG_OUTPUT_CHANGE() //PORTB^=BV(PB0)







//Display
///////////////////////////////

//Size
#define DISPLAY_X_WIDTH 12
#define DISPLAY_Y_HEIGHT 3

//LCD I2C Control      
#define LCD_DRIVER_ADDRESS (0x74)
//Control byte [Co,RS,R/_W,0,0,0,0,0]
#define LCD_CONTROL_BYTE_WRITE_COMMAND (0x80)
#define LCD_CONTROL_BYTE_WRITE_DATA    (0x40)
      
//LCD Setup
//[1,a6,a5,a4,a3,a2,a1,a0]
#define LCD_SET_DDRAM_ADDRESS(addr) (0x80 | addr)

//[0,1,a5,a4,a3,a2,a1,a0]
#define LCD_SET_CGRAM_ADDRESS(addr) (0x40 | addr)

//[0,0,1,4bit/_8Bit,N,M,0,0]
#define LCD_FUNCTION_SET      (0x34)

//[0,0,0,0,1,DisplayOn,CursorOn,BlinkOn]
#define LCD_DISPLAY_CONTROL   (0x0C)

//?
#define LCD_POWER_SAVE_SET    (0x01)

//[0,0,0,0,0,1,inc/_dec,shift_disp]
#define LCD_ENTRY_MODE        (0x06)

//[0,0,0,0,0,0,0,1]
#define LCD_CLEAR_DISPLAY (0x01)


#define LCD_GET_DDRAM_ADDR(x,y) ((x)+((y)<<5)+0x20)


//typedef unsigned char bool;


#endif //__HARDWARE_H__
