/*********************************************************************
 *   _      _        _     _                             _           *
 *  | |__  (_) _ __ | | __| |  ___  _ __    _ __    ___ | |_         *
 *  | '_ \ | || '__|| |/ /| | / _ \| '__|  | '_ \  / _ \| __|        *
 *  | |_) || || |   |   < | ||  __/| |   _ | | | ||  __/| |_         *
 *  |_.__/ |_||_|   |_|\_\|_| \___||_|  (_)|_| |_| \___| \__|        *
 *                                                                   *
 *********************************************************************
 * Copyright 2002                                                    *
 * Jrgen Birkler (birkler@yahoo.com)                                *
 *********************************************************************/

/**
 * @file
 * Timer library
 * @ingroup ATMEL_AVR_Embedded_TimerUtilities
 * @author Jrgen Birkler (birkler@yahoo.com) 
 */
                                                                  

/**
 * @defgroup ATMEL_AVR_Embedded_TimerUtilities Timers
 * Timer library, set reset and check expiry of timers.
 * @ingroup ATMEL_AVR_Embedded
 * 
 * Init library at CPU reset with #Timers_Init.
 *
 * Use #Timer_Set, #Timer_Reset, #Timer_HasExpired and 
 * #Timer_MillisecondsLeft in main program loop.
 *
 * Call #Timers_DecreaseAll from timer interrupt.
 * Resolution will be equal to the update value used
 * when updating the timers.
 */

/** 
 * \addtogroup ATMEL_AVR_Embedded_TimerUtilities
 * \@{
 */

#ifndef __TIMERS_H__
#define __TIMERS_H__

#include <stdbool.h>

///Timer value is in milliseconds
typedef unsigned short int Milliseconds_t;

///Id of an timer
typedef unsigned char TimerId_t;


//Include timers enum, enum must end with TIMER_LAST
#include "timer_ids.h"


/**
 * Init timer library, reset all timers
 * Must be called before any other timer functions are called
 */
void Timers_Init(void);

/**
 * Wait for timers to be updated before going to power save
 * Call this function before going to power save
 */
unsigned char Timers_GotoToIdle(void);


/**
 * Returns status of a timer
 * @param TimerId Id of timer
 * @retval !0  if timer has expired
 * @retval 0   if timer has not expired
 */
bool Timer_HasExpired(const TimerId_t TimerId);


/**
 * Resets a timer
 * Resets and cancels a timer
 * @param TimerId Id of timer
 */
void Timer_Reset(const TimerId_t TimerId);


/**
 * Starts timer
 * Starts timer, if timer is already running it will
 * be set with new timeout.
 * @param TimerId Id of timer
 * @param Timeout Milliseconds before timer expires
 */
void Timer_Set(const TimerId_t TimerId,const  Milliseconds_t Timeout);


/**
 * Returns milliseconds before timer expires
 * @param TimerId Id of timer
 * @returns Milliseconds left before timer expires,
 *          0 if timer is not running
 */
Milliseconds_t Timer_MillisecondsLeft(const TimerId_t TimerId);



/** \@} */

//extern void RTC_Increase(uint16_t milliseconds);

#endif //__TIMERS_H__
