/*********************************************************************
 *   _      _        _     _                             _           *
 *  | |__  (_) _ __ | | __| |  ___  _ __    _ __    ___ | |_         *
 *  | '_ \ | || '__|| |/ /| | / _ \| '__|  | '_ \  / _ \| __|        *
 *  | |_) || || |   |   < | ||  __/| |   _ | | | ||  __/| |_         *
 *  |_.__/ |_||_|   |_|\_\|_| \___||_|  (_)|_| |_| \___| \__|        *
 *                                                                   *
 *********************************************************************
 * Copyright 2002                                                    *
 * Jrgen Birkler (birkler@yahoo.com)                                *
 *********************************************************************/

/**
 * @file
 * FIR Filter
 * @ingroup ATMEL_AVR_Embedded_FIRFilter
 * @author Jrgen Birkler (birkler@yahoo.com) 
 */
                                                                  

/**
 * @defgroup ATMEL_AVR_Embedded_FIRFilter FIR Filter
 * FIR Filter functions
 *
 * @ingroup ATMEL_AVR_Embedded
 * 
 * See http://dspguru.com/info/faqs/fir/basics.htm
 *
 */


/** 
 * \addtogroup ATMEL_AVR_Embedded_FIRFilter
 * \@{
 */


#ifndef __FIR_FILTER_H__
#define __FIR_FILTER_H__

#include <inttypes.h>

///Size (number of taps) of a FIR filter
typedef uint8_t FIR_FilterSize_t;

///Sample
typedef int16_t FIR_FilterSample_t;

///Coefficient
typedef int8_t FIR_FilterCoefficient_t;

///Accumlated sum to avoid overrun this must be able to hold
typedef int32_t FIR_FilterAcculatedSum_t;

///Filter structure
typedef struct {
	FIR_FilterSize_t size;   ///<Number of taps
	FIR_FilterSize_t mask;       ///<Mask to position (2^log2size) - 1 
	FIR_FilterSize_t pos;        ///<Current insert position
	FIR_FilterAcculatedSum_t accumulatedSum; ///<Current accumulated sum
  FIR_FilterSample_t sum; ///<Sum
	FIR_FilterSample_t samples[1]; ///<Samples
} FIR_Filter_t;


/**
 * Defines a new filter variable
 * @param _name     Name of filter
 * @param _log2size Size of filter as log2 of actual size
 *
 * Use &_name.filter
 *
 * Example:
 * \code
 * #define myAverageLog2Size 3
 *
 * FIR_FILTER_DEFINE(myAverage,myAverageLog2Size);
 * 
 * FIR_Filter_Init(&myAverage.filter,myAverageLog2Size);
 * \endcode
 *
 */
#define FIR_FILTER_DEFINE(_name, _log2size) \
	typedef struct { \
		FIR_Filter_t filter; \
		FIR_FilterSample_t samples[1 << (_log2size)]; \
	} _name##_t; \
	static _name##_t _name; \



/**
 * Inits a FIR filter structure
 * @param filter_p  Pointer to filter structure
 * @param log2size  log2 of the size of the filter. Log2 is used to get
 *                  fast circular buffer, the
 *                  position can be masked by single &= instruction.
 */
void FIR_Filter_Init(FIR_Filter_t* filter_p, const FIR_FilterSize_t log2size);


/**
 * Inputs a sample to the filter using a BOXCAR coefficient table (all ones)
 *
 */
FIR_FilterSample_t FIR_Filter_BoxCar_Input(FIR_Filter_t* filter_p, const FIR_FilterSample_t sample);

/**
 * Returns the normalized sum of the filter. (Sum / taps * max_cooeffcient)
 *
 */
FIR_FilterSample_t FIR_Filter_GetNormalizedSum(const FIR_Filter_t* const filter_p);


/**@}*/



#endif //__FIR_FILTER_H__

