/*********************************************************************
 *   _      _        _     _                             _           *
 *  | |__  (_) _ __ | | __| |  ___  _ __    _ __    ___ | |_         *
 *  | '_ \ | || '__|| |/ /| | / _ \| '__|  | '_ \  / _ \| __|        *
 *  | |_) || || |   |   < | ||  __/| |   _ | | | ||  __/| |_         *
 *  |_.__/ |_||_|   |_|\_\|_| \___||_|  (_)|_| |_| \___| \__|        *
 *                                                                   *
 *********************************************************************
 * Copyright 2002                                                    *
 * Jrgen Birkler (birkler@yahoo.com)                                *
 *********************************************************************/

/**
 * @file
 * I2C implementation
 * @ingroup ATMEL_AVR_Embedded_I2C
 * @author Jrgen Birkler (birkler@yahoo.com) 
 */
                                                                  


#include "hardware.h"
#include "i2c.h"
#include "debug.h"


//Check defines in hardware.h
//////////////////////////////////////



#ifndef CLOCK_FREQ
#error CPU clock frequency not defined.
#define CLOCK_FREQ (40000000L) //Hz
#endif //CLOCK_FREQ

#ifndef I2C_RELEASE_SDA
#define I2C_RELEASE_SDA() {sbi(PORTC,PC5);cbi(DDRC,DDC5);}
#error I2C_RELEASE_SDE not defined.
#endif //I2C_RELEASE_SDA

#ifndef I2C_RELEASE_SCL
#error I2C_RELEASE_SCL not defined.
#define I2C_RELEASE_SCL() {sbi(PORTC,PC4);cbi(DDRC,DDC4);}
#endif //I2C_RELEASE_SCL

#ifndef I2C_DRIVE_SDA
#error I2C_DRIVE_SDA not defined.
#define I2C_DRIVE_SDA()   {sbi(DDRC,DDC5);cbi(PORTC,PC5);}
#endif //I2C_DRIVE_SDA

#ifndef I2C_DRIVE_SCL
#error I2C_DRIVE_SCL not defined.
#define I2C_DRIVE_SCL()   {sbi(DDRC,DDC4);cbi(PORTC,PC4);}
#endif //I2C_DRIVE_SCL

#ifndef I2C_IS_SDA_DRIVEN
#error I2C_IS_SDA_DRIVEN not defined.
#define I2C_IS_SDA_DRIVEN()    (bit_is_clear(PINC,PC5))
#endif //I2C_IS_SDA_DRIVEN


#ifndef I2C_IS_SCL_RELEASED
#error I2C_IS_SCL_RELEASED not defined.
#define I2C_IS_SCL_RELEASED()  (bit_is_set(PINC,PC4))
#endif //I2C_IS_SCL_RELEASED




//Some helper macros
////////////////////////////////////

#define I2C_SDA_INPUT() I2C_RELEASE_SDA()

#define I2C_INIT() { \
  I2C_RELEASE_SDA(); \
  I2C_RELEASE_SCL(); \
} \



/**
 * Generate bit delay
 * for a 3.3v device min clock pulse width low and high is 4 us
 */
static void I2C_BitDelay(void) 
{
  #define I2C_BIT_DELAY_uS (4L)
  #define I2C_CLOCKS_PER_LOOP (6L)
  #define I2C_BIT_COUNTS (I2C_BIT_DELAY_uS * CLOCK_FREQ / (1000000L * I2C_CLOCKS_PER_LOOP) )

  volatile uint8_t i = I2C_BIT_COUNTS;
  do {
  } while (i-->0);
}



/**
 * Release SLC and wait for it's release by slave
 * Wait max I2C_TIMER_COUNT bit delays
 * @retval >0 if slave released SCL (success).  
 * @retval 0 if slave did not release SCL before timer expired (fail)
 */
static uint8_t I2C_SCL_ReleaseAndWait(void) 
{
  #define I2C_TIMER_COUNT (200)

  uint8_t timer = I2C_TIMER_COUNT;
  I2C_RELEASE_SCL(); // release SCL
  // wait until SCL goes to a released state
  do
  {
    I2C_BitDelay();
  } while (!I2C_IS_SCL_RELEASED() && timer-- > 0);

  return timer;
}





void I2C_SendStartBit(void) 
{
  I2C_RELEASE_SDA();
  I2C_BitDelay();
  I2C_SCL_ReleaseAndWait();
  I2C_BitDelay();
  I2C_DRIVE_SDA();
  I2C_BitDelay();
  I2C_DRIVE_SCL();
  I2C_BitDelay();
}


bool I2C_SendByte(uint8_t byte) 
{
  bool acked;
  uint8_t i;

  for (i = 8; i-->0;) 
  {
    if (byte & 0x80)
    {
      I2C_RELEASE_SDA();
    }
    else 
    {
      I2C_DRIVE_SDA();
    }
      
    byte <<= 1; // send each bit, MSB first for address

    I2C_BitDelay();
    I2C_SCL_ReleaseAndWait();
    I2C_BitDelay();
    I2C_DRIVE_SCL();
    I2C_BitDelay();
  }

  // now for an ack
  I2C_RELEASE_SDA(); // release SDA ... listen for ACK
  I2C_SCL_ReleaseAndWait(); // master generates clock pulse for ACK
  I2C_BitDelay();  // delay so that slave have time to drive SDA

  // SDA at 0 ?
  acked = I2C_IS_SDA_DRIVEN(); 

  I2C_DRIVE_SCL();   //Back to business, ready for next byte or stop condition
  I2C_BitDelay();

  return acked;
}


/*
uint8_t I2C_GetByte(bool lastbyte) 
{
  // lastbyte == 1 for last byte
  uint8_t i;
  uint8_t result = 0;
   
  I2C_SDA_INPUT();

  for (i = 8;i-- > 0;)
  { 
    // each bit at a time, MSB first
    I2C_SCL_ReleaseAndWait();
    I2C_BitDelay();
    result<<=1;
    if (I2C_IS_SDA_RELEASED()) result |= 0x01;
    I2C_RELEASE_SCL();
    I2C_BitDelay();
  }
                       
  // send ACK
  // no ack on last byte ... lastbyte = 1 for the last byte
  if (lastbyte) 
  {
    //No ack
    I2C_RELEASE_SDA();
  }
  else
  {
    //Ack
    I2C_DRIVE_SDA();  
  }
  I2C_BitDelay();
  I2C_SCL_ReleaseAndWait();
  I2C_BitDelay();
  I2C_DRIVE_SCL();
  I2C_BitDelay();
  I2C_RELEASE_SDA();
  I2C_BitDelay();

  return(result);
}
*/

void I2C_SendStopBit(void) 
{
  I2C_DRIVE_SDA();
  I2C_BitDelay();
  I2C_SCL_ReleaseAndWait();
  I2C_BitDelay();
  I2C_RELEASE_SDA();
  I2C_BitDelay();
}

void I2C_Init(void)
{
  I2C_INIT();
}
