/*********************************************************************
 *   _      _        _     _                             _           *
 *  | |__  (_) _ __ | | __| |  ___  _ __    _ __    ___ | |_         *
 *  | '_ \ | || '__|| |/ /| | / _ \| '__|  | '_ \  / _ \| __|        *
 *  | |_) || || |   |   < | ||  __/| |   _ | | | ||  __/| |_         *
 *  |_.__/ |_||_|   |_|\_\|_| \___||_|  (_)|_| |_| \___| \__|        *
 *                                                                   *
 *********************************************************************
 * Copyright 2002                                                    *
 * Jrgen Birkler (birkler@yahoo.com)                                *
 *********************************************************************/


/**
 * @file
 *
 * @ingroup ATMEL_AVR_Embedded_I2C
 *
 * @author Jrgen Birkler (birkler@yahoo.com) 
 * @copyright 2002 Jrgen Birkler
 *
 */



/**
 * @defgroup ATMEL_AVR_Embedded_I2C I2C Interface
 *
 * Implementation for communicating with I2C slaves using
 * a polled I2C master implementation
 *
 * @ingroup ATMEL_EMBEDDED
 *
 * Handles I2C communications with external I2C devices.
 * Also called TWI (two wire interface), Find the i2c 
 * specification at http://www.philips.com.
 *
 * Devices are connected to two port pins on the chip.
 * The i2c bus uses two signals SCK - clock and SDA - Data.
 * Lines are active low and devices are ANDED together on the bus using 
 * open collector or open drain.
 *
 * The internal pull up resistors in the AVR are used so no external
 * pull up resistors are needed. Code has not been tested with multiple masters;
 * the bus arbitration is not tested. 
 *
 * Hardware configuration is done in hardware.h. The following macros are needed by 
 * i2c.c:
 * - I2C_RELEASE_SDA()     - Release the SDA line to a high state
 * - I2C_RELEASE_SCL()     - Release the SCL line to a high state
 * - I2C_DRIVE_SDA()       - Drive the SDA to a low state
 * - I2C_DRIVE_SCL()       - Drive the SCL to a low state
 * - I2C_IS_SDA_RELEASED() - Returns !0 if SDA line is released, 0 if SDA is driven (low)
 * - I2C_IS_SCL_RELEASED() - Returns !0 if SCL line is released, 0 if SCL is driven (low)
 *
 * First startbit, then slave address including LSB (read/write) then either
 * a byte send or byte read. Send bytes to LCD example:
 * \code
 *  I2C_SendStartBit();
 *	I2C_SendByte(LCD_DRIVER_ADDRESS); //I2C slave address
 *	I2C_SendByte(LCD_CONTROL_BYTE_WRITE_COMMAND);  //This address LCD control bytes
 *	I2C_SendByte(LCD_SET_DDRAM_ADDRESS(LCD_GET_DDRAM_ADDR(x,y)));
 *	I2C_SendStopBit();
 * \endcode
 *
 */



#ifndef __I2C_H__
#define __I2C_H__

#include <inttypes.h>
#include <stdbool.h>

/** 
 * \addtogroup ATMEL_AVR_Embedded_I2C
 * \@{
 */


///Initialize the I2C port pins
inline void I2C_Init(void);
 

/**
 * Send a start bit
 * Start bit is a 1 to 0 transition on SDA while SCL is high
 *
 * \code   
 *            ________
 * SDA master         \________
 *            _________________
 * SDA slave  ___/
 *            __________
 *        SCL ___/      \_____
 *
 * \endcode
 */
void I2C_SendStartBit(void);





/**
 * Send a byte (address or data) to the slave
 * parameter order used to select between sending LSB or MSB first
 *
 * @param byte Byte to send to slave
 *
 * @retval 1 if the slave did
 * @retval 0 if the slave didn't ack.
 * 
 * \code
 *            __________________________      ________
 * SDA slave                            \____/        ..
 *                                       ack
 *                ___ ___ ___   ___ __________ ___ ___
 * SDA master ___/___X___X___..X___X___/      \___x___..
 *    
 *                 _   _   _     _   _   _      _   _
 *       SCL  ____/1\_/2\_/3\.._/7\_/8\_/9\____/1\_/2\..
 *
 *                msb               lsb ack  ^
 *                                           |
 *                  [Can be held low by slave]       
 *                                     
 * \endcode
 *
 * @see I2C_SendStartBit
 * @see I2C_SendStopBit
 */
bool I2C_SendByte(uint8_t byte);


/**
 * Reads a byte from the slave
 *
 * @param lastbyte 1 if last byte read 0 if more bytes to be read
 * @returns byte read from slave
 *
 * \code
 *               _____________________      __________
 * SDA master __/                     \____/          ..
 *                                     ack
 *            _______ ___ ___   ___ ______________ ___
 * SDA slave      \__X___X___..X___X___/      \___x___..
 *
 *                 _   _   _     _   _   _      _   _
 *       SCL  ____/1\_/2\_/3\.._/7\_/8\_/9\____/1\_/2\..
 *
 *                msb               lsb ack  
 *                                     
 * \endcode
 */
uint8_t I2C_GetByte(bool lastbyte);



/**
 * Generates a stop bit
 * assumes SCL is low
 * Stop bit is a 0 to 1 transition on SDA while SCL is high
 *
 * \code   
 *           ____
 * SDA _____/    
 *         ______
 * SCL ___/      
 *    
 * \endcode
 * @see I2C_SendStartBit
 * @see I2C_SendByte
 *
 */
void I2C_SendStopBit(void);


/** \@} */


#endif //__I2C_H__
