/**
 @file
 
  Copyright (C) 1999 by Denis Chertykov (denisc@overta.ru)
  STK200 communication added by Tor Ringstad (torhr@pvv.ntnaddr.no).

  You can redistribute it and/or modify it under the terms of the GNU
  General Public License as published by the Free Software Foundation;
  either version 2, or (at your option) any later version.

  This program works with the connection scheme used in Atmels "STK200
  Starter Kit". The connection allows to program the AVR with Uros
  Platise `uisp' in `-dstk200' mode.

  Bugs: For me this work only if Parallell port in EPP mode.
        May be ECP mode also work?

  2002 June Jrgen Birkler (birkler@yahoo.com)
  - Updated code to work with GCC 3.2.
  - Renamed some of the communications functions to better reflect their origin
    'Monitor_'.
  - Added AT90S8535 as target.

  2003 April Jrgen Birkler (birkler@yahoo.com)
  - Using GCC 3.3
  - Updates to host-target protocol.
  - More comments to monitor trap loop
  - Trap loop now easier to configure to different hardware and interrupts
  - 8 Hardware breakpoints
  . 
  
  2003 November Jrgen Birkler (birkler@yahoo.com)
  - Reworked the protocol completetly, now using SPI
    - Hardware slave receives clock; no missed bits
    - MISO,MOSI,SCK to HOST. SS tied to MOSI.
    - RX bit is piggy backed in the SPI_SCK just before SS goes high
    - Uses SPI interrupt for assembler stepping
    .
  - Rewrote the PC monitor using C++ and threads
  .

*/

#include <stddef.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <inttypes.h>
#define AVR_MONITOR_TARGET
#include "monitor.h"

/*------------------------------------------------------------------*/


void monitor(void);


/*------------------------------------------------------------------*/
/*  Host communication via STK200 ISP                               */
/*------------------------------------------------------------------*/

///@todo More hardware configurations
#if (0 && defined(__AVR_AT90S2323__))
#elif  (__AVR_AT90S8515__ || __AVR_AT90S2313__  || __AVR_AT90S8535__)

#  define SPI_SCK PB7
#  define SPI_MISO PB6
#  define SPI_MOSI PB5
#  define SPI_SS PB4
#  define SPI_DDR DDRB
#  define SPI_PORT PORTB
#  define SPI_PIN PINB

#elif (__AVR_ATmega8__ || __AVR_AT90S4433__)

#  define SPI_MOSI PB3
#  define SPI_MISO PB4
#  define SPI_SCK PB5
#  define SPI_SS PB2
#  define SPI_DDR DDRB
#  define SPI_PORT PORTB
#  define SPI_PIN PINB
#  define INT0_DD    DDRD
#  define INT0_PORT  PORTD
#  define INT0_PIN   PD2
#  define INT1_DD    DDRD
#  define INT1_PORT  PORTD
#  define INT1_PIN   PD3
#else
#error "Don't know device type. Please define SPI pins."
#endif

#if defined(DOXYGEN)

///Define this to use timer0 for monitor,
///Timer0 cannot be used for any other purpose
#define AVR_MONITOR_TIMER0

///Define this to use external interrupt 0 for monitor
///Port must be free for monitor use, monitor will enable output on the
///corresponding port pin
#define AVR_MONITOR_INT0

///Define this to use external interrupt 1 for monitor
///Port must be free for monitor use, monitor will enable output on the
///corresponding port pin
#define AVR_MONITOR_INT1

#endif //defined(DOXYGEN)

///@todo jbirkler:Add Timer1 and Timer2 to supported interrupts
#if defined(AVR_MONITOR_TIMER0)
#define TIMER0_CK 1
#define MONITOR_INTERRUPT SIG_OVERFLOW0
#elif defined(AVR_MONITOR_INT0)
#if defined(INT0_DD) && defined(INT0_PORT) && defined(INT0_PIN)
#define MONITOR_INTERRUPT SIG_INTERRUPT0
#else
#error "Cannot use INT0. Define IO port for INT0"
#endif
#elif defined(AVR_MONITOR_INT1)
#if defined(INT1_DD) && defined(INT1_PORT) && defined(INT1_PIN)
#define MONITOR_INTERRUPT SIG_INTERRUPT1
#else
#error "Cannot use INT1. Define IO port for INT1"
#endif
#else
#define MONITOR_INTERRUPT SIG_SPI
#endif


#define SPI_ENABLE() \
  SPCR = _BV(SPE) | _BV(CPHA)

#define SPI_ENABLE_INT() \
  sbi(SPCR,SPIE)

#define SPI_DISABLE_INT() \
  sbi(SPCR,SPIE)

#define SPI_ENABLE_MASTER() \
  sbi(SPCR,MSTR)

#define SPI_WAIT_UNTIL_A_SLAVE() \
  loop_until_bit_is_clear(SPCR,MSTR) \

#define SPI_DISABLE() \
  SPCR = 0


/**
 * Enable for slave MISO is output, everything else is input with pull-up enabled
 * @author Jorgen Birkler (birkler@yahoo.com)
 * @hideinitializer
 */
#define SPI_SLAVE_INIT() { \
  sbi(SPI_PORT,SPI_SCK); \
  sbi(SPI_PORT,SPI_MOSI); \
  sbi(SPI_PORT,SPI_SS); \
  cbi(SPI_PORT,SPI_MISO); \
  cbi(SPI_DDR,SPI_SCK); \
  cbi(SPI_DDR,SPI_MOSI); \
  cbi(SPI_DDR,SPI_SS); \
  sbi(SPI_DDR,SPI_MISO); \
  SPI_ENABLE(); \
}

//#define _DEBUG

#ifdef _DEBUG

extern void dbg_sendchar(char data);
extern void dbg_sendbyte(unsigned char data);

#define D_(x) x

#else //#ifdef _DEBUG

#define D_(x)

#endif //#ifdef _DEBUG

#define SPI_TIMEOUT_COUNT 65000


void Monitor_InitComm(void)
{
  SPI_SLAVE_INIT();
}

/**
 * Sync host to target communication
 * with a trailing 10 ones followed by 4 zeros.
 * 
 * @author Jorgen Birkler (birkler@yahoo.com)
 *
 * 00111111111100xxxxxxxx
 */
//jbirkler:This will be six instructions
#define AVR_MONITOR_IS_RX_SYNC(val) ((((val)) & 0x00FFFF00) == 0x3FF000)
//jbirkler:Want this but the compiler (3.3.1) crashes:
//#define AVR_MONITOR_IS_RX_SYNC(val) ((((val>>8)) & 0x00FFFF) == 0x3FF0)


/**
 * Slave transmit
 * Wait until the previous tranfer is complete and writes to the data register
 *
 * @author Jorgen Birkler (birkler@yahoo.com)
 *
 * Its seems that the SPIF bit in SPSR is set after 8 bits have been shifted into
 * the shift register not when SS goes high. The datasheet doesn't explicitly say
 * but when reading the text you can get the impression that SS low-high triggers
 * the SPIF bit, this is not so.
 * The hardware seems also to be very confused if there is not exactly
 * 8 bits. 7 bits = nothing, 9 bits = garbage.
 *
 * By enabling clock phase CPHA (after testing I found that this works in mode 1
 * but not in mode 0) it is possible to piggy pack a
 * bit from the host to the target (remember that MOSI is controlling SS; 
 * MOSI cannot be used) by letting the SPI_SCK be either 1 or 0 when SS goes high.
 *
 * Bytes in the piggy back bit is synced so that the target can determine
 * when it received a complete byte, see #AVR_MONITOR_IS_RX_SYNC.
 *
 * @param tx_byte Byte to send
 * @retval -1 no byte recieved
 * @retval 0..255 byte recieved
 */
static int16_t SPI_Slave_Transfer(uint8_t tx_byte)
{
  uint32_t temp;
  int16_t retval = -1;
  
  //(Reset will set rx and timeout to zero, (see libc FAQ about startup init))
  static uint32_t rx;
  static uint16_t timeout;
  
  
  //Write to the transmit register
  SPDR = tx_byte;

  //Clear the transfer complete flag by reading data register
  if (SPDR)
    ;
    
  SPI_ENABLE();
  
  //Drive the MISO high to indicate that the AVR wants clock
  //If SS is low this has no effect until SS release because
  //if SS low the hardware has connected it to the output of the shift register.
  sbi(SPI_PORT,SPI_MISO);

  timeout = 0; 
  
  //Wait until the transfer is complete 
  //Wait for the SPIF bit and that SS is high
  while (bit_is_clear(SPSR,SPIF) || bit_is_clear(SPI_PIN,SPI_SS))
  {
    //if (timeout > SPI_TIMEOUT_COUNT) 
    //  continue;
    timeout++;
  }
    
  //We don't need the clock anymore
  cbi(SPI_PORT,SPI_MISO);
  SPI_DISABLE();
  
  //Transfer was successful, host seems to be alive, reset the timeout
  timeout = 0; 

  //Sent a byte and received a bit
  //Host sets SCK to zero or one before SS goes high
  temp = rx; //Help compiler with a temp variable
  temp <<=1;
  if (bit_is_set(SPI_PIN,SPI_SCK))
  {
    temp |= 0x1;
  }
    
  rx = temp;

  //Check if we have a complete received byte
  if (AVR_MONITOR_IS_RX_SYNC(rx))
  {
    //Returned the received character (host tx is more important...)
    return (rx&0x000000FF);
  }

  return retval;
}



/** 
 * High level routine for sending one byte of data. Blocks until the
 * host is ready to receive. 
 */
void Monitor_SendChar(uint8_t data)
{
  Monitor_InitComm();
  SPI_Slave_Transfer(data);
}


/**
 * High level routine for receiving one byte of data. Blocks until the
 * host is ready to send. 
 */
static uint8_t Monitor_ReceiveByte(void)
{
  int16_t temp;
  do {
    temp = SPI_Slave_Transfer(AVR_MONITOR_READY_TO_RECEIVE);
  } while (temp < 0);
  
  D_(dbg_sendbyte(temp);)
 
  return (temp);
} 



/*------------------------------------------------------------------*/
/*  Monitor code                                                    */
/*------------------------------------------------------------------*/


/* Extract high 8 bit of address */
#define hi8(x) \
 ({ unsigned char t; asm volatile ("ldi %0, hi8 (%1)" :"=r"(t) :"i"(x)); t;})
/* Extract low 8 bit of address */
#define lo8(x) \
 ({ unsigned char t; asm volatile ("ldi %0, lo8 (%1)" :"=r"(t) :"i"(x)); t;})

volatile avr_monitor_data_t avr_monitor_data;


void gdb_break(void);

#ifndef AVR_MONITOR_HW_BREAKPOINTS
#define AVR_MONITOR_HW_BREAKPOINTS (8)
#endif

#if (AVR_MONITOR_HW_BREAKPOINTS > 8)
#error "Max hw breakpoints is 8"
#endif

void monitor_enter(void) __attribute__ ((naked,weak));
void monitor_exit(void) __attribute__ ((naked,weak));


///@todo jbirkler:The function prototypes confuses the debugger a little. 
///Can this be avoided without breaking the const array?
void avr_monitor_hwbp_0(void) __attribute__ ((naked,weak));
void avr_monitor_hwbp_1(void) __attribute__ ((naked,weak));
void avr_monitor_hwbp_2(void) __attribute__ ((naked,weak));
void avr_monitor_hwbp_3(void) __attribute__ ((naked,weak));
#if (AVR_MONITOR_HW_BREAKPOINTS > 4)
void avr_monitor_hwbp_4(void) __attribute__ ((naked,weak));
void avr_monitor_hwbp_5(void) __attribute__ ((naked,weak));
void avr_monitor_hwbp_6(void) __attribute__ ((naked,weak));
void avr_monitor_hwbp_7(void) __attribute__ ((naked,weak));
#endif //#if (AVR_MONITOR_HW_BREAKPOINTS > 4)

///@todo jbirkler:Make hardware breakpoint size configurable for really die hard optimization
///Addresses to hardware breakpoints
static const avr_breakpoint_t PROGMEM hw_breakpoints[] = {
  (uint16_t)avr_monitor_hwbp_0,
  (uint16_t)avr_monitor_hwbp_1,
  (uint16_t)avr_monitor_hwbp_2,
  (uint16_t)avr_monitor_hwbp_3,
#if (AVR_MONITOR_HW_BREAKPOINTS > 4)
  (uint16_t)avr_monitor_hwbp_4,
  (uint16_t)avr_monitor_hwbp_5,
  (uint16_t)avr_monitor_hwbp_6,
  (uint16_t)avr_monitor_hwbp_7
#endif //#if (AVR_MONITOR_HW_BREAKPOINTS > 4)
};

static const avr_monitor_data_info_t PROGMEM monitor_info = {
  sizeof(avr_monitor_data_t),
  (avr_ptr_t)&avr_monitor_data,
  (avr_ptr_t)&hw_breakpoints
};



/* I don't use SIGNAL(SIG_OVERFLOW0) because it's a special function.
   __attribute__ ((naked)) seems better */


void MONITOR_INTERRUPT(void) __attribute__ ((naked));
void MONITOR_INTERRUPT(void)
{
  //jbirkler:routines called when entering and exiting the monitor
  asm volatile (".weak monitor_enter");
  asm volatile (".weak monitor_exit");

  //jbirkler:Creating weak references to the hardware breakpoint addresses
  //so that each hardware breakpoint symbols have an address
  //when linking even if not used.
    
  asm volatile (".weak avr_monitor_hwbp_0");
  asm volatile (".weak avr_monitor_hwbp_1");
  asm volatile (".weak avr_monitor_hwbp_2");
  asm volatile (".weak avr_monitor_hwbp_3");
#if (AVR_MONITOR_HW_BREAKPOINTS > 4)
  //Only weak if >4 otherwise not weak and user will get a 
  //reminder from the linker about a collision
  asm volatile (".weak avr_monitor_hwbp_4");
  asm volatile (".weak avr_monitor_hwbp_5");
  asm volatile (".weak avr_monitor_hwbp_6");
  asm volatile (".weak avr_monitor_hwbp_7");
#endif //#if (AVR_MONITOR_HW_BREAKPOINTS > 4)  
  asm volatile ("avr_monitor_hwbp_0:\n");
  asm volatile ("avr_monitor_hwbp_1:\n");
  asm volatile ("avr_monitor_hwbp_2:\n");
  asm volatile ("avr_monitor_hwbp_3:\n");
  asm volatile ("avr_monitor_hwbp_4:\n");
  asm volatile ("avr_monitor_hwbp_5:\n");
  asm volatile ("avr_monitor_hwbp_6:\n");
  asm volatile ("avr_monitor_hwbp_7:\n");
  asm volatile (".global  gdb_break\n");
  asm volatile ("gdb_break:\n");
    asm volatile ("push r27\n");             /* Temp save R27 */
    asm volatile ("in r27, __SREG__\n");     /* Save SREG */
    asm volatile ("cli\n");                  /* Disable interrupts. Needed for stepping. */

    asm volatile ("sts  avr_monitor_data+28, r28\n");    /* save r28 */
    asm volatile ("sts  avr_monitor_data+29, r29\n");    /* save r29 */
    asm volatile ("ldi  r28, lo8 (avr_monitor_data)\n"); /* y=r29:r28 = &regs */
    asm volatile ("ldi  r29, hi8 (avr_monitor_data)\n"); /* y=r29:r28 = &regs */
    asm volatile ("std  Y+32, r27\n");       /* Put SREG value to his place */
    asm volatile ("pop  r27\n");             /* Restore previous r27 value */
    asm volatile ("std  Y+27, r27\n");       /* and save it */
    asm volatile ("std  Y+30, r30\n");       /* save r30 */
    asm volatile ("std  Y+31, r31\n");       /* save r31 */
    asm volatile ("ldi  r31,0\n");           /* z = r31:r30 = 0 = &R0 */
    asm volatile ("ldi  r30,0\n");           /* z = r31:r30 = 0 = &R0 */
    asm volatile ("push_loop:\n");           /* save registers from 0 to 26 */
      asm volatile ("ld r27, Z+\n");  
      asm volatile ("st Y+, r27\n");
      asm volatile ("cpi  r30,27\n");
    asm volatile ("brne push_loop\n");

    asm volatile ("pop  r27\n");             /* pop high byte of ret address hi8(pm(pc)) */
    asm volatile ("pop  r26\n");             /* pop low byte of ret address lo8(pm(pc)) */
    asm volatile ("cpi  r27, hi8(pm(" "gdb_break" "))\n");
    asm volatile ("ldi  r24, lo8(pm(" "gdb_break" "))\n");
    asm volatile ("cpc  r26, r24\n");
    asm volatile ("brne __no_self_trace\n");
    asm volatile ("pop  r27\n");    /* pop high byte of ret address hi8(pm(pc)) */
    asm volatile ("pop  r26\n");    /* pop low byte of ret address lo8(pm(pc)) */

    asm volatile ("__no_self_trace:\n");
    asm volatile ("std  Y+35-27, r26\n");    // save return address (PC)
    asm volatile ("std  Y+36-27, r27\n");    // save return address (PC)
    asm volatile ("in r30, __SP_L__\n"::);   // get stack pointer (SP)
  #ifdef SPH
    asm volatile ("in r31, __SP_H__\n"::);    // get stack pointer (SP) 
  #else//#ifdef SPH
    asm volatile ("clr  r31");               // only 256 bytes of ram (no SPH) 
  #endif//#ifdef SPH
    asm volatile ("std  Y+33-27, r30\n");    // save SP 
    asm volatile ("std  Y+34-27, r31\n");    // save SP 
    asm volatile ("clr  __zero_reg__\n");
    asm volatile ("ldi  r30, lo8((avr_monitor_data+0x3F-1))"); // last byte in avr_monitor_data.stack 
    asm volatile ("out  __SP_L__, r30\n");   // setup internal monitor stack 
  #ifdef SPH
    asm volatile ("ldi  r31, hi8(avr_monitor_data+0x3F-1)\n"::);
    asm volatile ("out  __SP_H__, r31");
  #endif //#ifdef SPH

    SPI_DISABLE_INT();
    
    asm volatile ("rcall monitor_enter");
    
    ///@todo jbirkler:Make software breakpoint support optional. Even possible with GDB?
    ///@todo jbirkler:Inspect generated code and optimize loop by hand if necessary
    if (avr_monitor_data.continue_cmd == avr_monitor_step_to_next_bp)
    {
      unsigned char i = AVR_MONITOR_NUM_BREAKPOINTS;
      avr_breakpoint_t pc = avr_monitor_data.regs.pc;
      const avr_breakpoint_t* bp = (const avr_breakpoint_t*)&avr_monitor_data.breakpoint[0];

      //avr_breakpoint_t lastpc = avr_monitor_data.lastpc;
      //avr_monitor_data.lastpc = pc;
    
      do 
      {
        avr_breakpoint_t bp_address = *bp++;
        //if (lastpc < bp_address && bp_address <= pc) 
        if (bp_address == pc) 
        {
          //Yes, seems to have hit a breakpoint
          goto exec_mon;  // Bad style. but code quality more costly 
        }
      } while (--i > 0);

      //Nope, step next assembler intruction
      goto next_step;
    }

    exec_mon: 
      monitor();
      
    next_step:
      asm volatile ("rcall monitor_exit");

      asm volatile ("ldi  r31,0\n");           //r31:r30 = 0
      asm volatile ("ldi  r30,0\n");
      asm volatile ("ldi  r28, lo8(avr_monitor_data)\n"); //Y = avr_monitor_data
      asm volatile ("ldi  r29, hi8(avr_monitor_data)\n"); //.
      asm volatile ("pop_loop:\n");            //Restore register R0..R26
      asm volatile ("ld r27, Y+\n");           //.
      asm volatile ("st Z+, r27\n");           //.
      asm volatile ("cpi  r30, 27\n");         //.
      asm volatile ("brne pop_loop\n");        //Y now == 27
      asm volatile ("ldd  r30, Y+6\n");        //Restore SP
      asm volatile ("out  __SP_L__, r30\n");   //.
    #ifdef SPH
      asm volatile ("ldd  r31, Y+7\n");        //.
      asm volatile ("out  __SP_H__, r31\n");   //.
    #endif //#ifdef SPH
      asm volatile ("ldd  r27, Y+8\n");        //Restore PC by pushing on stack
      asm volatile ("push r27\n");             //.
      asm volatile ("ldd  r27, Y+9\n");        //.
      asm volatile ("push r27\n");             //.
      asm volatile ("ld r27, Y\n");            //.
      asm volatile ("ldd  r30, Y+3\n");        // real value of r30 pushed
      asm volatile ("push r30\n");             // .
      asm volatile ("ldd  r30, Y+5\n");        // r30 = regs.sreg 
      asm volatile ("lds  r31, avr_monitor_data+0x26\n");// r31 = .continue_cmd 
      asm volatile ("tst  r31\n");             // test .continue_cmd
      asm volatile ("ldd  r31, Y+4\n");        // Restore R31        
      asm volatile ("ldd  r28, Y+1\n");        // Restore R28
      asm volatile ("lds  r29, avr_monitor_data+29\n");       // Restore R29
      asm volatile ("brne monitor_stepping");  // .continue_cmd != 0 (avr_monitor_continue)
      asm volatile ("out  __SREG__, r30\n");   // Restore sreg
      asm volatile ("pop  r30\n");             // Restore R30
    //asm volatile ("monitor_enter:\n");
    //asm volatile ("monitor_exit:\n");
      asm volatile ("ret\n");                    // Go back

      asm volatile ("monitor_stepping:\n");

#if defined(AVR_MONITOR_TIMER0)

      asm volatile ("push r31");
      asm volatile ("in r31, %0\n"
                    "ori  r31, %1\n"
                    "out  %0, r31" :: "I" (_SFR_IO_ADDR(TIMSK)), "I" (BV (TOIE0)));
            /* set timer 0 counter */
      asm volatile ("ldi  r31,-4\n");
      asm volatile ("out  %0, r31" :: "I" (_SFR_IO_ADDR(TCNT0)));

      /* start timer 0 clocks */
      asm volatile ("ldi  r31, %0\n"
                    "out  %1, r31" :: "I"(TIMER0_CK), "I"(_SFR_IO_ADDR(TCCR0)));
      asm volatile ("pop  r31\n");

#elif defined(AVR_MONITOR_INT0)
      //Set INT0 Pin low
      sbi(INT0_DD,INT0_PIN);
      cbi(INT0_PORT,INT0_PIN);
      //Enable interrupt on INT0 low
      cbi(MCUCR,ISC01);
      cbi(MCUCR,ISC00);
      sbi(GIMSK,INT0);
#elif defined(AVR_MONITOR_INT1)
      //Set INT0 Pin low
      sbi(INT1_DD,INT1_PIN);
      cbi(INT1_PORT,INT1_PIN);
      //Enable interrupt on INT0 low
      cbi(MCUCR,ISC11);
      cbi(MCUCR,ISC10);
      sbi(GIMSK,INT1);
#else
      //jbirkler:If configured as a master and SS is low an interrupt is 
      //generated and the SPI switched over to be a slave
      
      SPI_ENABLE_INT();
      SPI_ENABLE_MASTER(); 
      //Request clock from host (don't really want it, just let host drive SS low)
      sbi(SPI_PORT,SPI_MISO);
      SPI_WAIT_UNTIL_A_SLAVE();
      cbi(SPI_PORT,SPI_MISO);

#endif

      asm volatile ("andi r30,0x7f\n"); /* clear I flag */
      asm volatile ("out  __SREG__, r30\n");
      asm volatile ("pop  r30\n");
      asm volatile ("reti");
}


static void monitor_send_byte(uint8_t b)
{
  Monitor_SendChar((b & 0xF) | 0x80);
  b >>= 4;
  Monitor_SendChar(b | 0x80);
}


#define monitor_send_address(lo,hi) \
  {monitor_send_byte(lo); \
  monitor_send_byte(hi);}
  

#define JOIN( X, Y ) DO_JOIN( X, Y )
#define DO_JOIN( X, Y ) DO_JOIN2(X,Y)
#define DO_JOIN2( X, Y ) X##Y


#define STATIC_ASSERT_(_pred) \
  typedef struct { char test[((!!(_pred)))*2-1];} JOIN(static_assert_typedef,__LINE__)
  




STATIC_ASSERT_(sizeof(avr_monitor_data.regs) == AVR_MONITOR_REGS_SIZE);  
STATIC_ASSERT_(offsetof(avr_monitor_data_t,hw_bp_flags) == AVR_MONITOR_BP_FLAGS_OFFSET);
STATIC_ASSERT_(offsetof(avr_monitor_data_t,continue_cmd) == AVR_MONITOR_CONTINUE_CMD_OFFSET);
STATIC_ASSERT_(offsetof(avr_monitor_data_t,breakpoint) == AVR_MONITOR_BREAKPOINTS_OFFSET);
STATIC_ASSERT_(sizeof(avr_monitor_data.breakpoint) == AVR_MONITOR_BREAKPOINTS_SIZE);
STATIC_ASSERT_(offsetof(avr_monitor_data_t,stack) == AVR_MONITOR_STACK_OFFSET);
STATIC_ASSERT_(sizeof(avr_monitor_data) == AVR_MONITOR_DATA_SIZE);
STATIC_ASSERT_(AVR_MONITOR_DATA_SIZE == 0x3F);
STATIC_ASSERT_(AVR_MONITOR_SIZEOF_DATA_INFO == sizeof(avr_monitor_data_info_t));


void monitor(void)
{
  ///@todo Inspect generated assembler for monitor loop and optimize
  for(;;)
  {
    uint8_t cmd;
    uint16_t addr;
    uint8_t data_or_length;
    
    D_(dbg_sendchar('<');)
    
    Monitor_SendChar(AVR_MONITOR_SIGTRAP);   // Send this byte for switching to monitor mode 

    // Receive command packet from host
    cmd = Monitor_ReceiveByte();
    addr = (Monitor_ReceiveByte() << 8 | 
            Monitor_ReceiveByte());
    data_or_length = Monitor_ReceiveByte();

    //Echo back
    Monitor_SendChar(cmd);
    monitor_send_address(addr>>8,addr&0xFF);
    monitor_send_byte(data_or_length);

    if (cmd == AVR_MONITOR_CMD_GET_MON_DATA)
    {
      addr = (uint16_t)&monitor_info;
      cmd = AVR_MONITOR_CMD_READ_FLASH;
    };
      
    //jbirkler:If statements smaller code than switch...
    if (cmd == AVR_MONITOR_CMD_READ_SRAM)
    {
      const uint8_t* p = (const uint8_t*)addr;
      //Make assumption here that length >0 (smaller code), host must assure
      do 
      {
        uint8_t data = *p++;
        monitor_send_byte(data);
        D_(dbg_sendbyte(data);)
      } while (--data_or_length > 0);
    }
    else if (cmd == AVR_MONITOR_CMD_READ_FLASH)    
    {
      do 
      {
        uint8_t data = PRG_RDB(addr++);
        monitor_send_byte(data);
        D_(dbg_sendbyte(data);)
      } while (--data_or_length > 0);
    }
    else if (cmd == AVR_MONITOR_CMD_WRITE_SRAM)    
    {
      *((uint8_t *)addr) = data_or_length;
    }

      
    Monitor_SendChar(0);
    
    if (cmd == AVR_MONITOR_CMD_EXIT)    
    {
      avr_monitor_data.continue_cmd = data_or_length;
      
      /* Quit (packet.data == 0) */
      /* Or Step (packet.data==1) */
      /* Or Stepping (packet.data == 2) */
      return;
    }
  } //for(;;))
}




