/**  
 * @file
 *
 * AVR Monitor for in System Debugging using MISO, MOSI and SCK
 * with host program that can connect to GDB using GDB serial protocol.
 *
 * Copyright (C) 1999 by Denis Chertykov (denisc@overta.ru)
 * Copyright (C) 2002 by Jrgen Birkler (jorgen@birkler.se)
 *
 * You can redistribute it and/or modify it under the terms of the GNU
 * General Public License as published by the Free Software Foundation;
 * either version 2, or (at your option) any later version.
 *
 * This program works with the connection scheme used in Atmels "STK200
 * Starter Kit". The connection allows to program the AVR with Uros
 * Platise `uisp' in `-dstk200' mode.
 *
 * 
 * @author Denis Chertykov (denisc@overta.ru)
 * @author Jrgen Birkler (jorgen@birkler.se)
 *
 * @par History
 * 1999 Denis Chertykov
 * - First version. Denis magic fingers write the trap interrupt loop
 * .
 * 2002 June Jrgen Birkler
 * - Updated code to work with GCC 3.2.
 * - Renamed some of the communications functions to better reflect their origin
 *   'Monitor_'.
 * - Added AT90S8535 as target.
 * .
 * 2003 April Jrgen Birkler
 * - Using GCC 3.3
 * - Updates to protocol. More comments to trap loop
 * - Trap loop now easier to configure to different hardware
 * - Added flashed breakpoints that are inserted compile time.
 * .
 *  2003 November Jrgen Birkler (jorgen@birkler.se)
 * - Reworked the protocol completetly, now using SPI
 *   - Can interrupt running code is SPI interrupt is enabled
 *   - Hardware slave receives clock, no missed bits
 *   - MISO,MOSI,SCK to HOST. SS tied to MOSI.
 *   .
 * - avr-monitor is now multi threaded and rewritten using C++
 *
 */

#ifndef _AVR_MONITOR_H__
#define _AVR_MONITOR_H__

#ifdef __cplusplus
extern "C" {
#endif


//extern void gdb_break(void);


/**
 * Send a char to the host
 * @param ch Character to send 1 <= ch <= 127 (may not be zero or above 127).
 */
extern void Monitor_SendChar(unsigned char ch);


/**
 * Trap and start the monitor
 * @hideinitializer
 */
#define TRAP() asm volatile ("rcall gdb_break")


/**
 * Creates a hardware breakpoint
 * @param num Must be unique number from 0..7
 * @hideinitializer
 */
#define BREAKPOINT(num) {\
  asm volatile (".global avr_monitor_hwbp_"#num"\n"); \
  asm volatile("avr_monitor_hwbp_"#num":"); \
  if (avr_monitor_data.hw_bp_flags & (0x1<<(num))) TRAP(); \
}



extern void monitor_enter(void);
extern void monitor_exit(void);


#if defined(AVR_MONITOR_HOST)
#include <sys/types.h>
//typedef unsigned short uint16_t;
//typedef signed short int16_t;
//typedef unsigned char uint8_t;
//typedef long unsigned int uint32_t;
#else
#include <inttypes.h>
#ifndef __AVR__
#error "Doesn't seem to be compiling for AVR."
#endif
#endif //defined(AVR_MONITOR_TARGET)


typedef uint16_t avr_ptr_t;
typedef uint16_t avr_sp_t;
typedef uint16_t avr_breakpoint_t;
typedef uint8_t avr_hw_bp_flag_t;


/** 
@defgroup avr_monitor_protocol Host-Target Protocol  

@section avr_monitor_protocol1 Hardware connections

The protocol uses the MISO, MOSI and SCK lines of the AVR. The AVR is setup as a SPI
slave and the PC is the SPI master. MOSI is tied to SS; PC controls the SS through the MOSI.
line.

\code
- PC -+                  +- AVR -
      |                  |
MISO  |------------<---- | MISO
      |                  |
MOSI  |------------>---- | MOSI
      |              |   |
      |              +-- | _SS
      |                  |
SCK   |------------>---- | SCK
------+                  +-------
\endcode

@section avr_monitor_protocol2 Low-level protocol

Transmit from Target (AVR/Slave) to Host (PC/Master)
-# Slave sets MISO high to indicate to master that it wants to send
-# Slave waits for a transfer complete (this will potentially send a zero to the host).
-# Slave reset MISO to high
-# Master drives SS low (by driving MOSI low). When SS is driven low the 
   MISO line is connected to the SP shift register
-# Clocks SCK
-# Reads bit from MISO before falling edge of SCK
-# Releases SS to high indicating to slave that transfer is complete
-# Since MOSI was driven low the SPI shift register contains zeros
.


Receive & transmit from target, setup is sck rising edge

\verbatim
                                7   6   5   4   3   2   1   0
                 ____________ ___ ___ ___ ___ ___ ___ ___ ___     
 MISO (trg)_____/          \_X___X___X___X___X___X___X___X___\___/\_________
                .                                                   .
           ________________                                       ____________ 
   SS (pc)                 \_____________________________________/  .   
                .          .                                     .  .   
           ______________  .  _   _   _   _   _   _   _   _   ________________  
  SCK (pc) ______________\___/ \_/ \_/ \_/ \_/ \_/ \_/ \_/ \_/________________ 
                .        . . . . . . . . . . . . . . . . . . .   .  .  .
                .....0....1.2.3.2.3.2.3.2.3.2.3.2.3.2.3.2.3.2..4....5...
               [A]        [B]                               [C] [D]
\endverbatim

- Note [A]: Target drives MISO low to request that a byte should be transferred
- Note [B]: When SS goes low avr hardware takes control of MISO
- Note [C]: After bit zero MISO will go to zero because zeros have been 
  shifted into the shift register.
- Note [D]: Target reads SCLK when SS goes high
- Note [D]: Target reads SCLK when SS goes high
.
- 0: Slave clock request response time
- 1: #TargetComm::SPI_SS_SETUP_DELAY
- 2: #TargetComm::SPI_SCK_LOW_DELAY
- 3: #TargetComm::SPI_SCK_HIGH_DELAY
- 4: #TargetComm::SPI_SCK_TX_DELAY or zero if no bit to transmit
- 5: #TargetComm::SPI_SS_RELEASE_DELAY or TargetComm::SPI_SS_TX_RELEASE_DELAY
.

@section avr_monitor_protocol3 Flow and byte level protocol
All bytes sent from the target are in the following ranges:
- [0x00] #AVR_MONITOR_NODATA
- [1..127] Debug trace information from application running in AVR
- [128..255] Data from debug monitor
.
- \ref AVR_MONITOR_READY_TO_RECEIVE
- \ref AVR_MONITOR_SIGTRAP
- \ref AVR_MONITOR_READY_TO_RECEIVE
- \ref avr_monitor_cmd_t
.
Each byte transmitted to the target is transmitted as 24 bits; each byte is 
by a 0011111111110000 header.

@section avr_monitor_protocol4 High level protocol

When the target is in running the monitor it sends a byte, 
#AVR_MONITOR_SIGTRAP. When this is read from the target by the host
switches to monitor mode and is ready to receive commands from the
debugger. 

Command packet host<->target

- \b AVR_MONITOR_CMD_READ_FLASH Read program memory
    - cmd = 0xC1 addr = [address to read] length = [bytes to read]
    - cmd = 0xC1 addr = [address read] length = [bytes read] buffer = [data]
    .
- \b AVR_MONITOR_CMD_READ_SRAM Read SRAM or IO area memory
    - cmd = 0xC2 addr = [address to read] length = [bytes to read]
    - cmd = 0xC2 addr = [address read] length = [bytes read] buffer = [data]
    .
- \b AVR_MONITOR_CMD_WRITE_SRAM Write SRAM or IO area  
    - cmd = 0xC3 addr = [address to write] data = [byte to write]
    - cmd = 0xC3 addr = [address written] length = [1] buffer = [byte written]
    .
- \b AVR_MONITOR_CMD_GET_MON_DATA Get monitor data (ram and flash addresses)
    - cmd = 0xC4 addr = [0x0000] length = [#AVR_MONITOR_SIZEOF_DATA_INFO]
    - cmd = 0xC4 addr = [0x0000] length = [#AVR_MONITOR_SIZEOF_DATA_INFO]
      buffer = [#avr_monitor_data_info_t]
    .  
- \b AVR_MONITOR_CMD_EXIT
    - cmd = 0xC0 addr = [0x0000] data = [#avr_monitor_run_mode_t]
    - cmd = 0xC0 addr = [0x0000] data = [#avr_monitor_run_mode_t]
*/
/**
 * @addtogroup avr_monitor_protocol 
 * \@{
 */

 
///Command to monitor in target
typedef enum { 
  ///Exit monitor. See #avr_monitor_run_mode_t
  AVR_MONITOR_CMD_EXIT   = 0xC0,  
  ///Read program memory
  AVR_MONITOR_CMD_READ_FLASH = 0xC1,
  ///Read SRAM
  AVR_MONITOR_CMD_READ_SRAM = 0xC2, 
  ///Write byte to SRAM
  AVR_MONITOR_CMD_WRITE_SRAM = 0xC3, 
  ///Get address of monitor data and flashed breakpoints
  AVR_MONITOR_CMD_GET_MON_DATA = 0xC4,   
  ///No operation
  AVR_MONITOR_CMD_NOP = 0xFF 
} avr_monitor_cmd_t;

///Monitor run mode
typedef enum {
  ///Go. Do not break on software breakpoints
	avr_monitor_continue = 0,             
  ///Step next asm instruction
	avr_monitor_step_asm_instruction = 1, 
  ///Step asm instructions until breakpoint
	avr_monitor_step_to_next_bp = 2       
} avr_monitor_run_mode_t;



///Character sent by target to indicate that it is switching to monitor mode
#define AVR_MONITOR_SIGTRAP 0xF7


///Sent by target to indicate that it is in receive mode
#define AVR_MONITOR_READY_TO_RECEIVE 0xFF


/**
 * Byte sent when monitor have no data to send
 * Must be zero since the shift register is filled with zeros after a transfer
 */
#define AVR_MONITOR_NODATA 0x00




/**\@}*/


/** 
 * @defgroup avr_monitor_data Monitor Data
 * Data structures used by the monitor to handle registers and
 * do communication.
 *
 * \@{
 */

#define AVR_GCC_IS_FLASH(addr)  ((addr) < 0x00800000) 
#define AVR_GCC_MASK_FLASH 0xFFFFF
#define AVR_GCC_IS_SRAMIO(addr) (0x00800000 <= (addr) && (addr) < 0x00810000)
#define AVR_GCC_MASK_SRAM 0xFFFF
#define AVR_GCC_IS_EEPROM(addr) (0x00810000 <= (addr) && (addr) < 0x00820000)
#define AVR_GCC_MASK_EEPROM 0xFFFF
 


/**
 * Registers in the avr. 
 *
 * It is used by:
 * - Target to store the registers when executing the monitor.
 * - GDB internally, it much match in AVR-GDB (avr-tdep.c) avr_register_raw_size(int regnum)
 *   avr_register_name (int regnum)
 * .
 *
 * @note Packed struct was not working for i386 gcc,
 *       avr-monitor.exe host does a manual translation
 */
typedef struct avr_registers_s {
  uint8_t reg[32];  ///<0: Registers r0-r31
  uint8_t sreg;     ///<32: Status register
  avr_sp_t sp;      ///<33-34: Stack Pointer
#if defined(AVR_MONITOR_TARGET)
  uint16_t pc;      ///<35-36:Program counter. (Word addresses), bytes address is PC*2
#elif defined(AVR_MONITOR_HOST)
  uint32_t pc;      ///<gdb uses byte addresses for PC, 32 bits
#endif
  ///0x25                    
} avr_registers_t;



///Number of software breakpoints
///@hideinitializer
#define AVR_MONITOR_NUM_BREAKPOINTS 4
///Size of the register structure
#define AVR_MONITOR_REGS_SIZE (0x25)
///Offset into the monitor data structure for breakpoints
#define AVR_MONITOR_BP_FLAGS_OFFSET AVR_MONITOR_REGS_SIZE
///Offset into the monitor data structure for the continue command
#define AVR_MONITOR_CONTINUE_CMD_OFFSET 0x26
///Offset into the monitor data structure for breakpoints
#define AVR_MONITOR_BREAKPOINTS_OFFSET (AVR_MONITOR_CONTINUE_CMD_OFFSET+1)
///Size of the breakpoint structure
#define AVR_MONITOR_BREAKPOINTS_SIZE (2*AVR_MONITOR_NUM_BREAKPOINTS)
///Not used yet. Offset in the monitor structure for the dedicated stack
#define AVR_MONITOR_STACK_OFFSET 0x2F
///Size of dedicated stack in bytes
#define AVR_MONITOR_STACK_SIZE 16
///Size of the whole #avr_monitor_data_t structure
#define AVR_MONITOR_DATA_SIZE (AVR_MONITOR_BREAKPOINTS_OFFSET + AVR_MONITOR_BREAKPOINTS_SIZE + AVR_MONITOR_STACK_SIZE)





/**
 * Data used by the target monitor to save registers
 * and to do communication
 *
 * Some advantages and disadvantages
 * - + Stack is separated from the application stack, if the application stack is corrupt
 *     the monitor can still run
 * - + The SP can be changed by the monitor
 *
 * - - The monitor save and restore block becomes a little more complicated (larger code)
 * .
 *
 * @todo jbirkler. One idea is to skip the regs in avr_monitor_data_t and only use a
 * dedicated stack that can hold all registers too. The monitor entry and exit block should
 * be a little simpler:
 * - Push to get some temp working regs
 * - Set SP to dedicated area
 * - Push everything to dedicated area
 * - Host knows in what order the monitor pushes things on the dedicated stack and can restore registers
 * .
 */
typedef struct avr_monitor_data_s {
  ///0x00: Saved registers
  avr_registers_t regs;
  
  ///0x25: Enabled hardware breakpoints, each bit correspond to a hw break point
  avr_hw_bp_flag_t hw_bp_flags;           
  
  ///0x26:
  uint8_t continue_cmd; ///<#avr_monitor_run_mode_t
  
  ///0x27: Software bp addresses. 0 if entry not used
  avr_breakpoint_t breakpoint[AVR_MONITOR_NUM_BREAKPOINTS]; 
  
  /** 0x2F:
   Make sure the monitor stack is large enough. When the size is
   changed here, remember to also change the value used in the inline
   assembly. 15 bytes is used right now.
   */
  uint8_t stack[AVR_MONITOR_STACK_SIZE];
  
  // 0x2F+16 = 0x3F = 63:
} avr_monitor_data_t;


/** 
 * Monitor data structure
 * Declared volatile so that the compiler doesn't optimize way a check to the 
 * hw_bp_flags.
 */
extern volatile avr_monitor_data_t avr_monitor_data;

#if defined(AVR_MONITOR_HOST) && !defined(PGM_VOID_P)
typedef avr_ptr_t PGM_VOID_P;
#endif 


///Monitor data info, see #
typedef struct {
  size_t size_of_monitor_data; ///sizeof(avr_monitor_data_t)
  avr_ptr_t data_address; ///address-of avr_monitor_data
  avr_ptr_t flashed_breakpoints_address; ///address of flashed breakpoints
} avr_monitor_data_info_t;


///sizeof(avr_monitor_data_info_t)
#define AVR_MONITOR_SIZEOF_DATA_INFO (6)

/**\@}*/


#ifdef __cplusplus
}
#endif


#endif //_AVR_MONITOR_H__


/**
 * @mainpage AVR Monitor
 *
 * \section avr_monitor_features Features
 * - In target debugging without JTAG interface
 *   - Enable disable flashed breakpoints and insert and remove software breakpoints
 *   - Step assembler instructions
 *   - View stack
 *   - Change register values
 *   .
 * - No extra wires besides the one used for serial programming (MISO,SCK and MOSI)
 * - Uses GDB serial protocol; no patches to GDB needed
 * - High speed debug tracing for application debug printouts
 * - Hot pluggable, can be connected to running target
 * - Multi threaded. GDB and user prompt available simultanously
 * - Caching of flash data and SRAM data for optimal performance
 * . 
 * 
 * \image html insight_screen_shot.png "AVR Insight debugging in an at90s4433"
 * 
 * \section avr_monitor_overview Overview
 * - \ref avr_monitor_intro Introduction to monitor and protocols
 * - \ref avr_monitor_building Building target, host and avr-gdb
 * - \ref avr_monitor_running Inspecting memory, setting breakpoints, running
 *
 * \section avr_monitor_intro Introduction
 * Most AVR devices do not have JTAG interface for in systems debugging.
 * With a small monitor in the target you can do in-system debugging.
 * The GNU debugger - GDB http://www.gnu.org/software/gnu and Insight 
 * http://www.redhat.com/sources/insight- have support for the AVR
 * thanks to the effort of Denis Chertykov and Theodore Roth.
 * 
 * \subsection avr_monitor_intro2 Connection
 * GDB connects to the AVR target through a proxy - the AVR host - using the GDB serial protocol.
 * The AVR Monitor host translates the 
 * \ref GDBThread "serial protocol commands" to the
 * AVR target protocol. The host and the target is connected through the MISO/MOSI/SCK
 * lines, so the monitor can be used if the ISP connector is available.
 * 
 * \code
 * [GDB] <- tcp/ip -> [AVR Host] <- MISO/MOSI/SCK -> [AVR Target]
 * \endcode
 *
 * \subsection avr_monitor_intro3 Target Monitor
 * The monitor uses an interrupt to do the assembler stepping and
 * to set breakpoints. Interrupt to use is configurable but default is to use SPI interrupt.
 * When a software breakpoints are set the monitor steps assembler
 * instructions one by one until a break point is reached.
 *
 * Assembler instructions are stepped by enabling and triggering an interrupt
 * in the monitor and then doing a RETI. When the AVR core returns from the interrupt
 * it will always execute one assembler instruction before the interrupt vector is loaded again.
 * 
 * The monitor cannot step instructions in interrupts that have higher priority
 * than the monitor or when interrupts are disabled. The monitor will break on flashed breakpoints
 * when interrupt is disabled, it is however not possible to step the code.
 *
 * \subsection avr_monitor_intro4 Breakpoints
 * The monitor can handle two different types of breakpoints. 
 *
 * - Flashed breakpoints are inserted during compilation with the #BREAKPOINT() macro. Each
 *   flashed breakpoint is accoscieted with a bit in the enabled flashed breakpoints mask. If the
 *   bit is set in the mask the monitor will be loaded when the BREAKPOINT() is reached.
 
 * - Software breakpoints can be inserted and removed during runtime. Software GDB inserts breakpoints
 *   when stepping out of functions, stepping into functions, etc. The number of software breakpoints
 *   are limited. If one or more software breakpoints are enabled the monitor steps assembler instructions 
 *   one by one until a breakpoint is reached. This is very slow.
 * .
 * when only hardware breakpoints are enabled the target can be run with full speed with only a very small
 * overhead at the compiled breakpoints.
 * All breakpoints are treated the same from GDB, the monitor host does the translation to a flashed breakpoint
 * if the address match one of the flashed breakpoints. The target monitor
 * supports eight flashed breakpoints.
 *
 * \section avr_monitor_building Configuring and Building
 * Download the host and target source code <a href="./avr-monitor.tar.gz">avr-monitor.tar.gz</a>
 *
 * \subsection avr_monitor_building_target Target
 * In the target you add monitor.c to your project. In source code you include monitor.h.
 *
 * You can select which interrupt to use for the monitor, currently SPI (default) INT0, INT1 or
 * TIMER0 is supported. It is fairly easy to add more. Look in monitor.c for available interrupt selections.
 *
 * Edit make; select correct MCU
 * \code
 * $ cd test
 * $ make flash
 * \endcode
 *
 * \subsection avr_monitor_building_host Host
 * \code
 * $ cd host
 * $ make
 * $ make install
 * \endcode
 * 
 * \subsection avr_monitor_building_gdb AVR Insight
 * Download snapshot or release later than 2003-05-05. Version 6.0 should be available.
 * \code
 * $ cd /usr/src
 * $ bunzip2 --keep --stdout insight-xxx.bz2 | tar --extract
 * $ mkdir avr-gdb
 * $ cd avr-gdb
 * $ /usr/src/insightxxx/configure --target=avr --program-prefix=avr- 
 * $ make
 * $ make install
 * \endcode
 * 
 * \section avr_monitor_running Running
 * Flash your target with the test program if not already done
 * \code
 * $ cd test
 * $ make flash
 * ...
 * \endcode
 * Open a new prompt and start the host, the host should report that it is waiting for GDB
 * \code
 * $ cd host
 * $ avr-monitor -d=2 --reset --gdb=localhost
 * \endcode
 * Start insight (gdb) and bring up the command prompt or use the menus.
 * Note: Make sure you don't have any spaces in the path for the .elf file,
 * Insight for CYGWIN cannot handle this, UNIX stuff I guess.
 * \code
 * $ cd test
 * $ avr-insight monitor_test.elf
 * \endcode
 * You should be able to see the debug information about the GDB commands set to the host.
 * Read flash, read SRAM, etc.
 *
 * \section avr_monitor_internals More information
 * - \ref avr_monitor_data
 * - \ref avr_monitor_protocol
 * .
 * - http://www.gnu.org/software/gdb
 * - http://sources.redhat.com/insight/
 * - http://www.birkler.se
 * - http://www.nongnu.org/avr-libc/
 * - http://winavr.sourceforge.net
 * .
 */



