/*********************************************************************
 *   _      _        _     _                             _           *
 *  | |__  (_) _ __ | | __| |  ___  _ __    _ __    ___ | |_         *
 *  | '_ \ | || '__|| |/ /| | / _ \| '__|  | '_ \  / _ \| __|        *
 *  | |_) || || |   |   < | ||  __/| |   _ | | | ||  __/| |_         *
 *  |_.__/ |_||_|   |_|\_\|_| \___||_|  (_)|_| |_| \___| \__|        *
 *                                                                   *
 *********************************************************************
 * Copyright 2002                                                    *
 * Jrgen Birkler (birkler@yahoo.com)                                *
 *********************************************************************/

/**
 * @file
 * Timer Utilities implementation
 * @ingroup ATMEL_AVR_Embedded_TimerUtilities
 * @author Jrgen Birkler (birkler@yahoo.com) 
 */
                                                                  

//#define DEBUG_ASSERT_ENABLE_ (1)

#include "timers.h"
#include "hardware.h"
#include "avr/signal.h"
#include "avr/interrupt.h"
#include "debug.h"
#include "rtc.h"

// **********************************************
// * Configuration
// **********************************************

/**
 * TIMER2_CLOCK_FREQ is set to the external crystal frequency in Hz, Usually 32768Hz.
 *
 * TIMER2_RTC_UPDATE_TIME should be set to the time between
 * updates to the RTC (Real Time Clock). Usually 1000 ms.
 * 
 * When using a crystal to clock the core the chip waits 35ms for the clock
 * to stabilize before any code is executed. The 35ms is default but can be changed
 * to 1ms if using a ceramic resonator by changing a fuse bit. See
 * the application note AN1259 about the RTC clock. 
 *
 * Larger updates times give lower power consumption but less accurate
 * timers. Lower values gives more accurate timers but higher power consumption.
 *
 * Accuracy of the RTC is only affected by the accuracy of the
 * crystal not by the selected TIMER2_RTC_UPDATE_TIME. The resolution
 * of the RTC is TIMER2_RTC_UPDATE_TIME/256.
 *
 * PRECISION/RESOLUTION/ACCURACY
 * 
 */
#ifndef TIMER2_CLOCK_FREQ
#define TIMER2_CLOCK_FREQ 32768 //Hz
#error "TIMER2_CLOCK_FREQ not defined in hardware.h. See config info."
#endif //TIMER2_CLOCK_FREQ

#ifndef TIMER2_RTC_UPDATE_TIME
#define TIMER2_RTC_UPDATE_TIME 1000 //ms
#error "TIMER2_RTC_UPDATE_TIME not defined in hardware.h. See config info."
#endif //TIMER2_RTC_UPDATE_TIME


///Prescaler requested by user
#define TIMER2_REQUESTED_PRESCALER ((TIMER2_CLOCK_FREQ) * 1000L / (256L * (TIMER2_RTC_UPDATE_TIME)))


//Now check the validity of the requested RTC update time
#if (256L * TIMER2_REQUESTED_PRESCALER * (TIMER2_RTC_UPDATE_TIME) != 1000L * (TIMER2_CLOCK_FREQ))
#error "Selected TIMER2_RTC_UPDATE_TIME not supported by crystal. See config info."
#endif //check update time

#if !defined(TIMER2_CLOCK_FREQ) || !defined(TIMER2_RTC_UPDATE_TIME) || !defined(TIMER2_REQUESTED_PRESCALER) 
#error "Requested prescaler not calculated"
#elif   (TIMER2_REQUESTED_PRESCALER == 1)
#define TCCR2_PRESCALE TCCR2_PRESCALE_1
#elif (TIMER2_REQUESTED_PRESCALER == 8)
#define TCCR2_PRESCALE TCCR2_PRESCALE_8
#elif (TIMER2_REQUESTED_PRESCALER == 32)
#define TCCR2_PRESCALE TCCR2_PRESCALE_32
#elif (TIMER2_REQUESTED_PRESCALER == 64)
#define TCCR2_PRESCALE TCCR2_PRESCALE_64
#elif (TIMER2_REQUESTED_PRESCALER == 128)
#define TCCR2_PRESCALE TCCR2_PRESCALE_128
#elif (TIMER2_REQUESTED_PRESCALER == 256)
#define TCCR2_PRESCALE TCCR2_PRESCALE_256
#elif (TIMER2_REQUESTED_PRESCALER == 1024)
#define TCCR2_PRESCALE TCCR2_PRESCALE_1024
#else
#error "Requested TIMER2_RTC_UPDATE_TIME not supported by TIMER2 prescaler. See config info."
#define TCCR2_PRESCALE TCCR2_PRESCALE_STOPPED
#endif




// **********************************************
// * Timers
// **********************************************


typedef struct {
  unsigned char expired;
  Milliseconds_t Value;
  Milliseconds_t left;
} Timer_t;

///All the timers
static volatile Timer_t Timers[TIMER_LAST];


///Timer is expired
#define TIMER_RESET (0)

#define TIMER_NO_UPDATE ((Milliseconds_t)(-1))

///Special type used by update function
typedef uint32_t LongMilliseconds_t;



// **********************************************
// * Hardware stuff
// **********************************************

#define TIMER2_ENABLE_OVERFLOW_INTERRUPT() sbi(TIMSK,TOIE2)
//Do not destroy the other flags (write zero to those)
#define TIMER2_CLEAR_COMPARE_FLAG() TIFR = BV(OCF2)
#define TIMER2_IS_COMPARE_FLAG_SET() bit_is_set(TIFR,OCF2)
#define TIMER2_DISABLE_OVERFLOW_INTERRUPT() cbi(TIMSK,TOIE2)
#define TIMER2_ENABLE_ASYNCHRONOUS() {sbi(ASSR,AS2);}
#define TIMER2_SET_COMPARE_VALUE(_val) OCR2 = (_val)
#define TIMER2_COUNTER() TCNT2
#define TIMER2_SET_COUNTER(_val) TCNT2 = (_val)
#define TIMER2_DISABLE_COMPARE_INTERRUPT() cbi(TIMSK,OCIE2)
#define TIMER2_ENABLE_COMPARE_INTERRUPT() sbi(TIMSK,OCIE2)
#define TIMER2_COMPARE_INTERRUPT_ENABLED() bit_is_set(TIMSK,OCIE2)
#define TIMER2_COMPARE_INTERRUPT_DISABLED() bit_is_clear(TIMSK,OCIE2)
#define TIMER2_ASYNC_SAFETY (1)
#define TIMER2_WAIT_ASYNC_UPDATE() { \
	uint8_t busy_flag;\
	do { \
		busy_flag = ASSR; \
		busy_flag &= BV(TCN2UB) | BV(OCR2UB) | BV(TCR2UB); \
	} while (busy_flag); }

#define TCCR2_CLEAR_ON_COMP_MATCH 0x80
#define TCCR2_COM_SET_OC2       BV(COM21) | BV(COM20)
#define TCCR2_COM_TOGGLE_OC2    BV(COM21)
#define TCCR2_COM_CLEAR_OC2     BV(COM20)
#define TCCR2_COM_DISCONNECT_OC2 0x00

#define TCCR2_PRESCALE_STOPPED 0x00
#define TCCR2_PRESCALE_1 0x01
#define TCCR2_PRESCALE_8 0x02
#define TCCR2_PRESCALE_32 0x03
#define TCCR2_PRESCALE_64 0x04
#define TCCR2_PRESCALE_128 0x05
#define TCCR2_PRESCALE_256 0x06
#define TCCR2_PRESCALE_1024 0x07

 

/**
 * Updates the timers by inspecting the current timer2 counter
 * and calculating elapsed time. 
 */
static inline void Timers_Update(void);


/**
 * Decreases all timers with delta ms
 * @param delta milliseconds elapsed
 * @returns Milliseconds until next update or TIMER_NO_UPDATE if not timers
 * are running.
 */
static Milliseconds_t Timers_DecreaseAll(const Milliseconds_t delta);



//TIMER2 Interrupts begin

///The output match compare interrupt
INTERRUPT(SIG_OUTPUT_COMPARE2)
{
	Timers_Update();
}

///Timer2 overflow interrupt
INTERRUPT(SIG_OVERFLOW2)
{
  //TIMER2_DEBUG_OUTPUT_CHANGE();
	//DEBUG_PRINT_INFO_(".");
	RTC_Increase(TIMER2_RTC_UPDATE_TIME);
}

//TIMER2 Interrupts end

void Timers_Init(void)
{
	//TIMER2_DEBUG_OUTPUT_INIT();
	TIMER2_DISABLE_COMPARE_INTERRUPT(); 
	TIMER2_DISABLE_OVERFLOW_INTERRUPT(); 
	TIMER2_ENABLE_ASYNCHRONOUS(); 
  TCCR2 =  TCCR2_COM_DISCONNECT_OC2 | TCCR2_PRESCALE; 
	TIMER2_SET_COUNTER(0); 
	TIMER2_SET_COMPARE_VALUE(0); 
	{
		TimerId_t id;
		for (id=0;id < TIMER_LAST;id++)
		{
      Timers[id].Value=TIMER_RESET;
      //Timers[id].expired = 0;
		}
	}
	TIMER2_WAIT_ASYNC_UPDATE(); 
	TIMER2_ENABLE_OVERFLOW_INTERRUPT(); 
}


unsigned char Timers_GotoToIdle(void)
{
	TimerId_t id;
	TIMER2_WAIT_ASYNC_UPDATE();
	for (id=0;id < TIMER_LAST;id++)
	{
		if (Timers[id].expired > 0)
		{
			return 0;
		}
	}
	return 1;
}





/**
 * Update the timers
 * -# Calculate ticks elapsed since last update
 * -# Calculate ms elapsed since last update
 * -# Update all timers and get ms until next update
 * -# Calculate ticks until next update
 * -# Enable interrupt
 */
static void Timers_Update(void)
{
	static uint8_t previousTimer2Count = 0;
	static uint8_t subUpdateDelta = 0; //256th's of a TIMER2_RTC_UPDATE_TIME

	//Disable compare match interrupt and clear interrupt flag
	uint8_t interruptEnabled = TIMER2_COMPARE_INTERRUPT_ENABLED();
	TIMER2_DISABLE_COMPARE_INTERRUPT();
	
	uint8_t outputCompareMatch = TIMER2_IS_COMPARE_FLAG_SET();
	TIMER2_CLEAR_COMPARE_FLAG();
	//We should now be safe from TIMER2 interrupts...
	
	uint8_t currentTimer2Count = TIMER2_COUNTER();
	LongMilliseconds_t updateDelta;

	//Interupt enabled?
	if (interruptEnabled)
	{
		uint8_t ticksElapsed;
		//Yes, the compare match is set
		ticksElapsed = 0;

		//Calculate ticks elapsed since previous update, 
		//take into account that it's a 8bit counter
		ticksElapsed += currentTimer2Count;
		ticksElapsed -= previousTimer2Count;

		//dbg_printf("c=%d p=%d elapsed=%d ",(int)currentTimer2Count,(int)previousTimer2Count,(int)ticksElapsed);
		//DEBUG_ASSERT_(ticksElapsed < 110);

		//Calculate ms elapsed
		if (outputCompareMatch && ticksElapsed == 0)
		{
			updateDelta = 256L * TIMER2_RTC_UPDATE_TIME;
		}
		else {
			updateDelta = ticksElapsed;
			updateDelta *= TIMER2_RTC_UPDATE_TIME;
		}

		updateDelta += subUpdateDelta;

		//ms = updateDelta / 256
		//subMs =  updateDelta % 256
		//But don't use the compiler...
		//We MUST count the sub ms also, otherwise the small timer will never be decreased
		subUpdateDelta = (uint8_t)updateDelta;
		updateDelta /= 256;
	}
	else {
		//No, compare counter is not running.
		updateDelta = 0;
	}

	//DEBUG_DECIMAL_(updateDelta);
	updateDelta = Timers_DecreaseAll((Milliseconds_t)updateDelta);

  //DEBUG_DECIMAL_(updateDelta);

	//Any timers?
	if ((Milliseconds_t)updateDelta != TIMER_NO_UPDATE)
	{
		//Yes, timers are running.
		DEBUG_ASSERT_(updateDelta > 0);

		//compareValue to be set in TIMER2 compare match register
		uint8_t timer2CompareMatch;

		//Synchronize the compare match interrupt to the
		//overflow interrupt if the update is longer than TIMER2_RTC_UPDATE_TIME
		//by setting compare match to 0.

		//Do we need to count ticks
		if ((Milliseconds_t)updateDelta < TIMER2_RTC_UPDATE_TIME)
		{
			//Yes, we need to count ticks

			//ticks = 256 * updateDelta / TIMER2_RTC_UPDATE_TIME

			//There are 256 ticks to every one TIMER2_RTC_UPDATE_TIME
			updateDelta *= 256; //Should translate to <<=8 (it does!)

			updateDelta /= TIMER2_RTC_UPDATE_TIME;

			DEBUG_ASSERT_(updateDelta <= 255);

			/*
			if ((uint8_t)updateDelta > 100)
				timer2CompareMatch = 100;
			else */
				timer2CompareMatch = updateDelta; 

			//Calculate TIMER2 ticks until update
			/*
			while (updateDelta >= TIMER2_RTC_UPDATE_TIME && 
						timer2CompareMatch < (155-TIMER2_ASYNC_SAFETY))
			{
				updateDelta -= TIMER2_RTC_UPDATE_TIME;
				timer2CompareMatch++;
			}*/

			//Interrupt will only trigger if the counter is counted to
			// the compare match. Add one.
			timer2CompareMatch++;

			timer2CompareMatch+=currentTimer2Count;

		}
		else {
			//No, don't need to count ticks this time
			//set compare match count to 0 to sync with overflow interrupt
			timer2CompareMatch = 0;
		}

		//Set new compare value based on new timers
		//Enable interrupt
		TIMER2_SET_COMPARE_VALUE(timer2CompareMatch);

		//Store the current counter as the previous
		previousTimer2Count = currentTimer2Count;

		//New compare value set, enable interrupt
		//Do we need to wait for update of async registers?
		//Hope not (assume not), we may be executing in an interrupt here...
		TIMER2_ENABLE_COMPARE_INTERRUPT();
	}
	else {
		//No timers running, do not enable compare interrupt
		//DEBUG_PRINT_INFO_("\r\nNo timers");
	}
}



static Milliseconds_t Timers_DecreaseAll(const Milliseconds_t delta)
{
	TimerId_t id;
	Milliseconds_t nextTime = TIMER_NO_UPDATE;
	for (id=0; id < TIMER_LAST;id++)
	{
    if (Timers[id].Value != TIMER_RESET)
		{
			if (delta >= Timers[id].left) 
			{
        Timers[id].left += Timers[id].Value;
        Timers[id].expired++;
      }
			Timers[id].left -= delta;

      if (Timers[id].left < nextTime)
      {
        nextTime = Timers[id].left;
      }
		}
	}
	return nextTime;
}



bool Timer_HasExpired(const TimerId_t TimerId)
{
  cli();
  if (Timers[TimerId].expired > 0)
  {
    Timers[TimerId].expired--;
    sei();
    return true;
  }
  else {
    sei();
    return false;
  }
}

void Timer_Reset(const TimerId_t TimerId)
{
  cli();
  Timers[TimerId].Value=TIMER_RESET;
  Timers[TimerId].left=TIMER_RESET;
  Timers[TimerId].expired=0;
  sei();
}


void Timer_Set(const TimerId_t TimerId,const Milliseconds_t setTimeout)
{
	Timers_Update();
  cli();
  Timers[TimerId].Value=setTimeout;
  Timers[TimerId].left=setTimeout;
  Timers[TimerId].expired=0;
  sei();
	Timers_Update();
}

void Timer_SetExpired(const TimerId_t TimerId,const Milliseconds_t setTimeout)
{
	Timers_Update();
  cli();
  Timers[TimerId].Value=setTimeout;
  Timers[TimerId].left=0;
  Timers[TimerId].expired=0;
  sei();
	Timers_Update();
}



/*
Milliseconds_t Timer_GetMillisecondsLeft(const TimerId_t TimerId)
{
	Milliseconds_t result;
	cli();
  if (Timers[TimerId] == TIMER_RESET)
  {
    result = 0;
  }
  else {
		result = Timers[TimerId];
  }
	sei();

	return result;
}
*/



